from typing import Sequence, Union

from bluesky import plan_stubs as bps
from ophyd import Device

from ..devices.shutter import ShutterState
from ..instrument import beamline


def _set_shutters(shutters: Union[str, Sequence[Device]], direction: int):
    if shutters != []:
        shutters = beamline.registry.findall(shutters)
    # Prepare the plan
    plan_args = [obj for shutter in shutters for obj in (shutter, direction)]
    if len(plan_args) > 0:
        # Emit the messages
        yield from bps.mv(*plan_args)


def open_shutters(shutters: Union[str, Sequence[Device]]):
    """A plan to open the shutters.

    By default, this plan is greedy and will open all shutters defined
    at the beamline. If only specific shutters should be opened, they
    can be passed either as Device objects or device names as the
    *shutters* argument.

    E.g.
      RE(open_shutters(["endstation_shutter"]))

    E.g.
      shutter = haven.devices.shutter.Shutter(..., name="Shutter C")
      RE(open_shutters([shutter]))

    This plan will temporarily remove the default shutter-related
    suspenders from the haven run engine.

    """
    yield from _set_shutters(shutters, ShutterState.OPEN)


def close_shutters(shutters: Union[str, Sequence[Device]] = "shutters"):
    """A plan to close some shutters.

    By default, this plan is lazy and requires any shutters to be
    passed explicitly. Shutters can be passed either as Device objects
    or device names as the *shutters* argument.

    E.g.
      RE(close_shutters(["Shutter C"]))

    E.g.
      shutter = haven.devices.shutter.Shutter(..., name="Shutter C")
      RE(close_shutters([shutter]))

    This plan will temporarily remove the default shutter-related
    suspenders from the haven run engine.

    """
    yield from _set_shutters(shutters, ShutterState.CLOSED)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from bluesky.callbacks import CallbackBase
from tqdm import tqdm


class ProgressBar(CallbackBase):
    num_points: int = 0

    """A progress bar that tracks a scan."""

    def start(self, doc):
        self.num_points = doc["num_points"]
        self.plan_name = doc.get("plan_name", "Scanning")
        # Set up the tqdm bar
        self.pbar = tqdm(total=self.num_points, unit="pt", desc=self.plan_name)

    def event(self, doc):
        self.pbar.update(1)

    def stop(self, doc):
        self.pbar.close()


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from ophydregistry.exceptions import (  # noqa: F401
    ComponentNotFound,
    InvalidComponentLabel,
    MultipleComponentsFound,
)


class NoEnergies(ValueError):
    """The plan parameters do not result in any energies being selected for scanning."""

    ...


class GainOverflow(RuntimeError):
    """The gain is trying to be set to a value that is not allowed."""

    ...


class FileNotWritable(IOError):
    """Output file is available but does not have write intent."""

    ...


class XDIFilenameKeyNotFound(KeyError):
    """The format string for an XDI filename has unknown placeholders."""

    ...


class InvalidPV(ValueError):
    """A process variable path or similar is expected, but the path
    provided is not valid.

    """

    ...


class DocumentNotFound(RuntimeError):
    """An attempt was made to use a document, but the requested document
    was not available.

    """

    ...


class IOCTimeout(RuntimeError):
    """The IOC did not start within the allotted time."""

    ...


class InvalidConfiguration(TypeError):
    """The configuration files for Haven are missing keys."""

    ...


class UnknownDeviceConfiguration(InvalidConfiguration):
    """The configuration for a device does not match the known options."""

    ...


class InvalidHarmonic(ValueError):
    """The requested harmonic is invalid for this insertion device."""

    ...


class SignalNotFound(KeyError):
    """The dataset is not present in the run."""

    ...


class EmptySignalName(ValueError): ...


class InvalidTransformation(TypeError):
    """The data cannot be transformed to a new y-data signal."""

    ...


class InvalidScanParameters(ValueError):
    """The given scan parameters will not produce a sane scan."""

    ...


class PluginNotPrimed(RuntimeError):
    """The detector plugin has not yet received a dataframe.

    Prior to starting capture for some detector file writer plugins,
    the plugin need to receive at least one dataframe in order to
    extract the dimensions, type, etc. of the data it will receive.

    """

    ...


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest
from bluesky_queueserver_api import BPlan
from qtpy import QtCore

from firefly.energy import EnergyDisplay
from haven.devices.energy_positioner import EnergyPositioner


@pytest.fixture()
async def energy_positioner(sim_registry):
    energy = EnergyPositioner(
        monochromator_prefix="mono_ioc:",
        undulator_prefix="id_ioc:",
        name="energy",
    )
    await energy.connect(mock=True)
    sim_registry.register(energy)
    return energy


@pytest.fixture()
def display(qtbot, energy_positioner):
    # Load display
    display = EnergyDisplay()
    qtbot.addWidget(display)
    return display


async def test_move_energy(qtbot, display):
    # Click the set energy button
    btn = display.ui.set_energy_button
    expected_item = BPlan("set_energy", energy=8402.0)

    def check_item(item):
        return item.to_dict() == expected_item.to_dict()

    qtbot.keyClicks(display.target_energy_lineedit, "8402")
    with qtbot.waitSignal(
        display.queue_item_submitted, timeout=1000, check_params_cb=check_item
    ):
        qtbot.mouseClick(btn, QtCore.Qt.LeftButton)


def test_predefined_energies(qtbot, display):
    # Check that the combo box was populated
    combo_box = display.ui.edge_combo_box
    assert combo_box.count() > 0
    assert combo_box.itemText(0) == "Select edge…"
    assert combo_box.itemText(1) == "Ca K (4038 eV)"
    # Does it filter energies outside the usable range?
    assert combo_box.count() < 250
    # Does it update the energy line edit?
    with qtbot.waitSignal(combo_box.activated, timeout=1000):
        qtbot.keyClicks(combo_box, "Ni K (8333 eV)\t")
        combo_box.activated.emit(9)  # <- this shouldn't be necessary
    line_edit = display.ui.target_energy_lineedit
    assert line_edit.text() == "8333.000"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import os
from pathlib import Path

import numpy as np
import pandas as pd

# from pydm.data_plugins import plugin_modules, add_plugin
import pytest
from ophyd import DynamicDeviceComponent as DCpt
from ophyd import Kind
from ophyd.sim import instantiate_fake_device, make_fake_device
from tiled.adapters.mapping import MapAdapter
from tiled.adapters.xarray import DatasetAdapter
from tiled.client import Context, from_context
from tiled.server.app import build_app

import haven
from haven._iconfig import beamline_connected as _beamline_connected
from haven.catalog import Catalog
from haven.devices.aps import ApsMachine
from haven.devices.area_detector import AravisDetector
from haven.devices.beamline_manager import BeamlineManager, IOCManager
from haven.devices.dxp import DxpDetector
from haven.devices.dxp import add_mcas as add_dxp_mcas
from haven.devices.ion_chamber import IonChamber
from haven.devices.monochromator import Monochromator
from haven.devices.robot import Robot
from haven.devices.shutter import PssShutter
from haven.devices.slits import ApertureSlits, BladeSlits
from haven.devices.xia_pfcu import PFCUFilter, PFCUFilterBank, PFCUShutter
from haven.devices.xspress import Xspress3Detector
from haven.devices.xspress import add_mcas as add_xspress_mcas

top_dir = Path(__file__).parent.resolve()
haven_dir = top_dir / "haven"


# Specify the configuration files to use for testing
os.environ["HAVEN_CONFIG_FILES"] = ",".join(
    [
        f"{haven_dir/'iconfig_testing.toml'}",
        f"{haven_dir/'iconfig_default.toml'}",
    ]
)


@pytest.fixture()
def beamline_connected():
    with _beamline_connected(True):
        yield


@pytest.fixture()
def sim_registry(monkeypatch):
    # Save the registry so we can restore it later
    registry = haven.beamline.registry
    objects_by_name = registry._objects_by_name
    objects_by_label = registry._objects_by_label
    registry.clear()
    # Run the test
    try:
        yield registry
    finally:
        # Restore the previous registry components
        registry.clear()
        registry._objects_by_name = objects_by_name
        registry._objects_by_label = objects_by_label


@pytest.fixture()
async def ion_chamber(sim_registry):
    ion_chamber = IonChamber(
        scaler_prefix="255idcVME:3820:",
        scaler_channel=2,
        preamp_prefix="255idc:SR03",
        voltmeter_prefix="255idc:LJT7_Voltmeter0:",
        voltmeter_channel=1,
        counts_per_volt_second=10e6,
        name="I00",
    )
    # Connect to the ion chamber
    await ion_chamber.connect(mock=True)
    sim_registry.register(ion_chamber)
    return ion_chamber


@pytest.fixture()
def blade_slits(sim_registry):
    """A fake set of slits using the 4-blade setup."""
    FakeSlits = make_fake_device(BladeSlits)
    slits = FakeSlits(prefix="255idc:KB_slits", name="kb_slits", labels={"slits"})
    sim_registry.register(slits)
    return slits


class SimpleBeamlineManager(BeamlineManager):
    """For a fake class, we need to un-override __new__ to just make
    itself.

    """

    iocs = DCpt(
        {
            "ioc255idb": (IOCManager, "ioc255idb:", {}),
            "ioc255idc": (IOCManager, "ioc255idc:", {}),
        }
    )

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)


@pytest.fixture()
def beamline_manager(sim_registry):
    """A fake set of slits using the 4-blade setup."""
    FakeManager = make_fake_device(SimpleBeamlineManager)
    manager = FakeManager(
        prefix="companionCube:", name="companion_cube", labels={"beamline_manager"}
    )
    sim_registry.register(manager)
    return manager


@pytest.fixture()
def aperture_slits(sim_registry):
    """A fake slit assembling using the rotary aperture design."""
    FakeSlits = make_fake_device(ApertureSlits)
    slits = FakeSlits(
        prefix="255ida:slits:US:",
        name="whitebeam_slits",
        pitch_motor="m3",
        yaw_motor="m4",
        horizontal_motor="m1",
        diagonal_motor="m2",
        labels={"slits"},
    )
    sim_registry.register(slits)
    return slits


@pytest.fixture()
def sim_camera(sim_registry):
    FakeCamera = make_fake_device(AravisDetector)
    camera = FakeCamera(name="s255id-gige-A", labels={"cameras", "area_detectors"})
    sim_registry.register(camera)
    camera.pva.pv_name._readback = "255idSimDet:Pva1:Image"
    # Registry with the simulated registry
    yield camera


class DxpVortex(DxpDetector):
    mcas = DCpt(
        add_dxp_mcas(range_=[0, 1, 2, 3]),
        kind=Kind.normal | Kind.hinted,
        default_read_attrs=[f"mca{i}" for i in [0, 1, 2, 3]],
        default_configuration_attrs=[f"mca{i}" for i in [0, 1, 2, 3]],
    )


@pytest.fixture()
def dxp(sim_registry):
    FakeDXP = make_fake_device(DxpVortex)
    vortex = FakeDXP(name="vortex_me4", labels={"xrf_detectors", "detectors"})
    # vortex.net_cdf.dimensions.set([1477326, 1, 1])
    sim_registry.register(vortex)
    yield vortex


class Xspress3Vortex(Xspress3Detector):
    mcas = DCpt(
        add_xspress_mcas(range_=[0, 1, 2, 3]),
        kind=Kind.normal | Kind.hinted,
        default_read_attrs=[f"mca{i}" for i in [0, 1, 2, 3]],
        default_configuration_attrs=[f"mca{i}" for i in [0, 1, 2, 3]],
    )


@pytest.fixture()
def xspress(sim_registry):
    FakeXspress = make_fake_device(Xspress3Vortex)
    vortex = FakeXspress(name="vortex_me4", labels={"xrf_detectors"})
    sim_registry.register(vortex)
    yield vortex


@pytest.fixture()
def robot(sim_registry):
    RobotClass = make_fake_device(Robot)
    robot = RobotClass(name="robotA", prefix="255idA:")
    sim_registry.register(robot)
    return robot


@pytest.fixture()
def mono(sim_registry):
    mono = instantiate_fake_device(Monochromator, name="monochromator")
    sim_registry.register(mono)
    yield mono


@pytest.fixture()
def aps(sim_registry):
    aps = instantiate_fake_device(ApsMachine, name="APS")
    sim_registry.register(aps)
    yield aps


@pytest.fixture()
def xia_shutter_bank(sim_registry):
    class ShutterBank(PFCUFilterBank):
        shutters = DCpt(
            {
                "shutter_0": (
                    PFCUShutter,
                    "",
                    {"top_filter": 4, "bottom_filter": 3, "labels": {"shutters"}},
                )
            }
        )

        def __new__(cls, *args, **kwargs):
            return object.__new__(cls)

    FakeBank = make_fake_device(ShutterBank)
    bank = FakeBank(prefix="255id:pfcu4:", name="xia_filter_bank", shutters=[[3, 4]])
    sim_registry.register(bank)
    yield bank


@pytest.fixture()
def xia_shutter(xia_shutter_bank):
    shutter = xia_shutter_bank.shutters.shutter_0
    yield shutter


@pytest.fixture()
def shutters(sim_registry):
    FakeShutter = make_fake_device(PssShutter)
    kw = dict(
        prefix="_prefix",
        labels={"shutters"},
    )
    shutters = [
        FakeShutter(name="Shutter A", **kw),
        FakeShutter(name="Shutter C", **kw),
    ]
    [sim_registry.register(s) for s in shutters]
    yield shutters


@pytest.fixture()
def filters(sim_registry):
    FakeFilter = make_fake_device(PFCUFilter)
    kw = {
        "labels": {"filters"},
    }
    filters = [
        FakeFilter(name="Filter A", prefix="filter1", **kw),
        FakeFilter(name="Filter B", prefix="filter2", **kw),
    ]
    [sim_registry.register(f) for f in filters]
    return filters


# Tiled data to use for testing
# Some mocked test data
run1 = pd.DataFrame(
    {
        "energy_energy": np.linspace(8300, 8400, num=100),
        "It_net_counts": np.abs(np.sin(np.linspace(0, 4 * np.pi, num=100))),
        "I0_net_counts": np.linspace(1, 2, num=100),
    }
).to_xarray()

grid_scan = pd.DataFrame(
    {
        "CdnIPreKb": np.linspace(0, 104, num=105),
        "It_net_counts": np.linspace(0, 104, num=105),
        "aerotech_horiz": np.linspace(0, 104, num=105),
        "aerotech_vert": np.linspace(0, 104, num=105),
    }
).to_xarray()

hints = {
    "energy": {"fields": ["energy_energy", "energy_id_energy_readback"]},
}

bluesky_mapping = {
    "7d1daf1d-60c7-4aa7-a668-d1cd97e5335f": MapAdapter(
        {
            "primary": MapAdapter(
                {
                    "data": DatasetAdapter.from_dataset(run1),
                },
                metadata={"descriptors": [{"hints": hints}]},
            ),
        },
        metadata={
            "plan_name": "xafs_scan",
            "start": {
                "plan_name": "xafs_scan",
                "uid": "7d1daf1d-60c7-4aa7-a668-d1cd97e5335f",
                "hints": {"dimensions": [[["energy_energy"], "primary"]]},
            },
        },
    ),
    "9d33bf66-9701-4ee3-90f4-3be730bc226c": MapAdapter(
        {
            "primary": MapAdapter(
                {
                    "data": DatasetAdapter.from_dataset(run1),
                },
                metadata={"descriptors": [{"hints": hints}]},
            ),
        },
        metadata={
            "start": {
                "plan_name": "rel_scan",
                "uid": "9d33bf66-9701-4ee3-90f4-3be730bc226c",
                "hints": {"dimensions": [[["pitch2"], "primary"]]},
            }
        },
    ),
    # 2D grid scan map data
    "85573831-f4b4-4f64-b613-a6007bf03a8d": MapAdapter(
        {
            "primary": MapAdapter(
                {
                    "data": DatasetAdapter.from_dataset(grid_scan),
                },
                metadata={
                    "descriptors": [
                        {
                            "hints": {
                                "Ipreslit": {"fields": ["Ipreslit_net_counts"]},
                                "CdnIPreKb": {"fields": ["CdnIPreKb_net_counts"]},
                                "I0": {"fields": ["I0_net_counts"]},
                                "CdnIt": {"fields": ["CdnIt_net_counts"]},
                                "aerotech_vert": {"fields": ["aerotech_vert"]},
                                "aerotech_horiz": {"fields": ["aerotech_horiz"]},
                                "Ipre_KB": {"fields": ["Ipre_KB_net_counts"]},
                                "CdnI0": {"fields": ["CdnI0_net_counts"]},
                                "It": {"fields": ["It_net_counts"]},
                            }
                        }
                    ]
                },
            ),
        },
        metadata={
            "start": {
                "plan_name": "grid_scan",
                "uid": "85573831-f4b4-4f64-b613-a6007bf03a8d",
                "hints": {
                    "dimensions": [
                        [["aerotech_vert"], "primary"],
                        [["aerotech_horiz"], "primary"],
                    ],
                    "gridding": "rectilinear",
                },
                "shape": [5, 21],
                "extents": [[-80, 80], [-100, 100]],
            },
        },
    ),
}


mapping = {
    "255id_testing": MapAdapter(bluesky_mapping),
}

tree = MapAdapter(mapping)


@pytest.fixture(scope="session")
def tiled_client():
    app = build_app(tree)
    with Context.from_app(app) as context:
        client = from_context(context)
        yield client["255id_testing"]


@pytest.fixture(scope="session")
def catalog(tiled_client):
    return Catalog(client=tiled_client)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest
from qtpy.QtCore import Qt
from qtpy.QtGui import QStandardItemModel

from firefly.bss import BssDisplay


@pytest.fixture()
def bss_api(mocker):
    api = mocker.MagicMock()
    api.listESAFs.return_value = [
        {
            "description": (
                "We will perform some K-edge and L-edge XAFS measurements of "
                "some transition metal nanoparticle powder samples such as "
                "silver, palladium, gold, copper and platinum. \r\n"
                "\r\n"
                "Some of the measurements will need in situ gas adsorption "
                "and/or heating conditions. Standard beamline temperature "
                "controller and power supply will be used to heat samples to "
                "100C. Some of the samples may be in solution phase. We will "
                "also perform Au L2 edge XAFS measurement in HERFD mode of "
                "some metal nanoparticle samples. The nanoparticle samples "
                "are all unbound nanostructured materials. Samples will be "
                "encapsulated in Kapton tape or prepared by using capillaries "
                "and quartz wool. For solution phase measurements, the Teflon "
                "sample holder will be used."
            ),
            "esafId": 269238,
            "esafStatus": "Pending",
            "esafTitle": (
                "A Partner User Proposal to Continue the Successful "
                "Collaboration between the Canadian Light Source Inc. and the "
                "Advanced Photon Source"
            ),
            "experimentEndDate": "2023-03-31 08:00:00",
            "experimentStartDate": "2023-03-28 08:00:00",
            "experimentUsers": [
                {
                    "badge": "86423",
                    "badgeNumber": "86423",
                    "email": "peng.zhang@dal.ca",
                    "firstName": "Peng",
                    "lastName": "Zhang",
                    "piFlag": "Yes",
                },
                {
                    "badge": "302308",
                    "badgeNumber": "302308",
                    "email": "yannachen@anl.gov",
                    "firstName": "Yanna",
                    "lastName": "Chen",
                    "piFlag": "No",
                },
                {
                    "badge": "299574",
                    "badgeNumber": "299574",
                    "email": "dmeira@anl.gov",
                    "firstName": "Debora",
                    "lastName": "Motta Meira",
                    "piFlag": "No",
                },
                {
                    "badge": "300051",
                    "badgeNumber": "300051",
                    "email": "zy916874@dal.ca",
                    "firstName": "Ziyi",
                    "lastName": "Chen",
                    "piFlag": "No",
                },
            ],
            "sector": "25",
        }
    ]

    api.listProposals.return_value = [
        {
            "title": (
                "A Partner User Proposal to Continue the Successful Collaboration"
                " between the Canadian Light Source Inc. and the Advanced Photon Source"
            ),
            "id": 74163,
            "experimenters": [
                {
                    "id": 521867,
                    "badge": "307373",
                    "email": "gianluigi.botton@lightsource.ca",
                    "piFlag": "Y",
                    "instId": 3949,
                    "firstName": "Gianluigi",
                    "lastName": "Botton",
                    "institution": "Canadian Light Source",
                },
                {
                    "id": 488536,
                    "badge": "242561",
                    "email": "chithra.karunakaran@lightsource.ca",
                    "instId": 3949,
                    "firstName": "Chithra",
                    "lastName": "Karunakaran",
                    "institution": "Canadian Light Source",
                },
                {
                    "id": 488508,
                    "badge": "299574",
                    "email": "dmeira@anl.gov",
                    "instId": 3949,
                    "firstName": "Debora",
                    "lastName": "Motta Meira",
                    "institution": "Canadian Light Source",
                },
            ],
            "activities": [
                {
                    "startTime": "2023-02-14 08:00:00-06:00",
                    "endTime": "2023-02-17 08:00:00-06:00",
                    "duration": 259200,
                },
                {
                    "startTime": "2023-03-28 08:00:00-05:00",
                    "endTime": "2023-03-31 08:00:00-05:00",
                    "duration": 259200,
                },
            ],
            "submittedDate": "2021-03-04 13:20:06-06:00",
            "proprietaryFlag": "N",
            "startTime": "2023-02-14 08:00:00-06:00",
            "endTime": "2023-03-31 08:00:00-05:00",
            "duration": 518400,
            "cycle": "2023-1",
        }
    ]
    yield api


@pytest.fixture()
def bss(beamline_manager):
    bss = beamline_manager.bss
    bss.proposal.mail_in_flag.sim_put(1)
    bss.proposal.proprietary_flag.sim_put(1)
    return bss


@pytest.fixture()
def display(qtbot, bss_api, bss):
    display = BssDisplay(api=bss_api)
    qtbot.addWidget(display)
    return display


def test_bss_proposal_model(display):
    assert display.ui_filename() == "bss.ui"
    # Check model construction
    assert isinstance(display.proposal_model, QStandardItemModel)
    assert display.proposal_model.rowCount() > 0
    # Check that the view has the model attached
    assert display.ui.proposal_view.model() is display.proposal_model


def test_bss_proposal_updating(display, qtbot, bss):
    # Set some base-line values on the IOC
    bss.proposal.proposal_id.set("").wait()
    # Change the proposal item
    item = display.proposal_model.item(0, 1)
    assert item is not None
    rect = display.proposal_view.visualRect(item.index())
    # See if the proposal PVs were updated
    with qtbot.waitSignal(display.proposal_selected):
        qtbot.mouseClick(
            display.proposal_view.viewport(), Qt.LeftButton, pos=rect.center()
        )
    assert display.ui.update_proposal_button.isEnabled()
    pv_id = bss.proposal.proposal_id.get(use_monitor=False)
    assert pv_id != "74163"
    assert display._proposal_id == "74163"
    # Now update the PROPOSAL details
    with qtbot.waitSignal(display.proposal_changed):
        qtbot.mouseClick(display.ui.update_proposal_button, Qt.LeftButton)
    pv_id = bss.proposal.proposal_id.get(use_monitor=False)
    assert pv_id == "74163"


def test_bss_proposals(display, bss_api):
    # Check values
    api_proposal = bss_api.listProposals()[0]
    proposals = display.proposals()
    proposal = proposals[0]
    assert proposal["Title"] == api_proposal["title"]
    assert proposal["ID"] == api_proposal["id"]
    assert proposal["Users"] == "Botton, Karunakaran, Motta Meira"
    assert proposal["Badges"] == "307373, 242561, 299574"
    assert proposal["Start"] == "2023-02-14 08:00:00-06:00"
    assert proposal["End"] == "2023-03-31 08:00:00-05:00"


def test_bss_esaf_model(bss_api, bss, qtbot):
    bss.proposal.beamline_name.set("255-ID-Z").wait()
    bss.esaf.aps_cycle.set("3024-1").wait()
    display = BssDisplay(api=bss_api)
    qtbot.addWidget(display)
    assert display.ui_filename() == "bss.ui"
    # Check model construction
    assert isinstance(display.esaf_model, QStandardItemModel)
    # assert display.esaf_model.rowCount() > 0
    bss_api.listESAFs.assert_called_once_with("3024-1", "255")
    # Check that the view has the model attached
    assert display.ui.esaf_view.model() is display.esaf_model


def test_bss_esaf_updating(display, qtbot, bss):
    # Set some base-line values on the IOC
    bss.esaf.esaf_id.set("").wait()
    # Change the ESAF item
    item = display.esaf_model.item(0, 1)
    assert item is not None
    rect = display.esaf_view.visualRect(item.index())
    # Clicking a list entry should enable the update button
    with qtbot.waitSignal(display.esaf_selected):
        qtbot.mouseClick(display.esaf_view.viewport(), Qt.LeftButton, pos=rect.center())
    assert display.ui.update_esaf_button.isEnabled()
    pv_id = bss.esaf.esaf_id.get(use_monitor=False)
    assert pv_id != "269238"
    assert display._esaf_id == "269238"
    # Now update the ESAF details
    with qtbot.waitSignal(display.esaf_changed):
        qtbot.mouseClick(display.ui.update_esaf_button, Qt.LeftButton)
    pv_id = bss.esaf.esaf_id.get(use_monitor=False)
    assert pv_id == "269238"


def test_bss_esafs(display, bss_api):
    # Check values
    api_esaf = bss_api.listESAFs()[0]
    esafs = display.esafs
    esaf = esafs()[0]
    assert esaf["Title"] == api_esaf["esafTitle"]
    assert esaf["ID"] == api_esaf["esafId"]
    assert esaf["Users"] == "Zhang, Chen, Motta Meira, Chen"
    assert esaf["Badges"] == "86423, 302308, 299574, 300051"
    assert esaf["Start"] == "2023-03-28 08:00:00"
    assert esaf["End"] == "2023-03-31 08:00:00"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest
from qtpy.QtWidgets import QAction

from firefly.controller import ActionsRegistry
from firefly.main_window import FireflyMainWindow, PlanMainWindow


@pytest.fixture()
def window(qapp, qtbot):
    window = FireflyMainWindow()
    qtbot.addWidget(window)
    return window


@pytest.fixture()
def actions():
    registry = ActionsRegistry()
    return registry


def test_navbar(qapp, qtbot, actions):
    actions.queue_controls = {
        "pause": QAction(),
        "start": QAction(),
    }
    window = PlanMainWindow(actions=actions)
    qtbot.addWidget(window)
    # Check that the navbar actions are set up properly
    assert hasattr(window.ui, "navbar")
    navbar = window.ui.navbar
    # Navigation actions are removed
    assert window.ui.actionHome not in navbar.actions()
    # Run engine actions have been added to the navbar
    assert actions.queue_controls["pause"] in navbar.actions()
    assert actions.queue_controls["start"] in navbar.actions()


def test_navbar_autohide(qtbot, actions):
    """Test that the queue navbar is only visible when plans are queued."""
    window = PlanMainWindow(actions=actions)
    qtbot.addWidget(window)
    window.show()
    navbar = window.ui.navbar
    # Pretend the queue has some things in it
    window.update_queue_controls({"in_use": True})
    assert navbar.isVisible()
    # Make the queue be empty
    window.update_queue_controls({"in_use": False})
    assert not navbar.isVisible()


def test_add_menu_action(window):
    # Check that it's not set up with the right menu yet
    assert not hasattr(window, "actionMake_Salad")
    # Add a menu item
    action = window.add_menu_action(
        action_name="actionMake_Salad", text="Make Salad", menu=window.ui.menuTools
    )
    assert hasattr(window.ui, "actionMake_Salad")
    assert hasattr(window, "actionMake_Salad")
    assert action.text() == "Make Salad"
    assert action.objectName() == "actionMake_Salad"


def test_motor_menu(qapp, qtbot, actions):
    actions.motors = {
        "A": QAction(),
        "B": QAction(),
        "C": QAction(),
    }
    window = FireflyMainWindow(actions=actions)
    qtbot.addWidget(window)
    # Check that the menu items have been created
    assert hasattr(window.ui, "positioners_menu")
    assert hasattr(window.ui, "motors_menu")
    assert window.ui.motors_menu.actions() == list(actions.motors.values())


def test_plans_menu(qapp, qtbot, actions):
    actions.plans = {
        "start": QAction(),
        "pause": QAction(),
        "abort": QAction(),
    }
    window = FireflyMainWindow(actions=actions)
    qtbot.addWidget(window)
    assert hasattr(window.ui, "plans_menu")
    window.ui.plans_menu.actions() == actions.plans


def test_show_message(window):
    window.statusBar()
    # Send a message
    window.show_status("Hello, APS.")


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
import time
from datetime import datetime
from typing import Mapping, Sequence

import intake
from bluesky import plan_stubs as bps
from bluesky import plans as bp
from pydantic import BaseModel
from rich import print as rprint
from tiled.queries import Key, Regex

from .catalog import Catalog, tiled_client
from .instrument import beamline

log = logging.getLogger(__name__)


__all__ = [
    "save_motor_position",
    "list_motor_positions",
    "recall_motor_position",
    "list_current_motor_positions",
]


class MotorAxis(BaseModel):
    name: str
    readback: float
    offset: float | None = None

    def as_dict(self):
        return {"name": self.name, "readback": self.readback, "offset": self.offset}


class MotorPosition(BaseModel):
    name: str
    motors: Sequence[MotorAxis]
    uid: str | None = None
    savetime: float | None = None

    @classmethod
    def _load(Cls, run_md, data_keys, data):
        """Common routines for synch and async loading."""
        if run_md["start"]["plan_name"] != "save_motor_position":
            raise ValueError(f"Run {run_md['start']['uid']} is not a motor position.")
        # Extract motor positions from the run
        motor_axes = []
        for data_key in data_keys.values():
            mname = data_key["object_name"]
            axis = MotorAxis(
                name=mname,
                readback=data[mname][0],
            )
            motor_axes.append(axis)
        # Create the motor position object
        return Cls(
            name=run_md["start"]["position_name"],
            motors=motor_axes,
            uid=run_md["start"]["uid"],
            savetime=run_md["start"]["time"],
        )

    @classmethod
    def load(Cls, run):
        """Create a new MotorPosition object from a Tiled Bluesky run."""
        return Cls._load(
            run_md=run.metadata,
            # Assumes the 0-th descriptor is for the primary stream
            data_keys=run["primary"].metadata["descriptors"][0]["data_keys"],
            data=run["primary"]["data"].read(),
        )

    @classmethod
    async def aload(Cls, scan):
        """Create a new MotorPosition object from a Tiled Bluesky run.
        Similar to ``MotorPosition.load()``, but asynchronous."""
        return Cls._load(
            run_md=await scan.metadata,
            data_keys=await scan.data_keys(),
            data=await scan.data(),
        )


def default_collection():
    catalog = intake.catalog.load_combo_catalog()["haven"]
    client = catalog._asset_registry_db.client
    collection = client.get_database().get_collection("motor_positions")
    return collection


# Prepare the motor positions
async def rbv(motor):
    """Helper function to get readback value (rbv)."""
    if hasattr(motor, "readback"):
        return await motor.readback.get_value()
    elif hasattr(motor, "user_readback"):
        return await motor.user_readback.get_value()
    else:
        return await motor.get_value()


def save_motor_position(*motors, name: str, md: Mapping = {}):
    """A Bluesky plan to Save the current positions of a number of motors.

    Parameters
    ==========
    *motors
      The list of motors (or motor names/labels) whose position to
      save.
    name
      A human-readable name for this position (e.g. "sample center")
    md
      Additional metadata to store with the motor position.

    """
    # Resolve device names or labels
    motors = beamline.registry.findall(motors)
    # Create the new run object
    _md = {
        "position_name": name,
        "plan_name": "save_motor_position",
    }
    _md.update(md)
    yield from bp.count(motors, md=_md)


def print_motor_position(position):
    # Prepare metadata strings for the header
    metadata = []
    if position.uid is not None:
        metadata.append(f'uid="{position.uid}"')
    if position.savetime is not None:
        timestamp = datetime.fromtimestamp(position.savetime).strftime(
            "%Y-%m-%d %H:%M:%S"
        )
        ts_str = f"{timestamp}"
        metadata.append(ts_str)
    if len(metadata) > 0:
        metadata_str = f"{', '.join(metadata)}"
    else:
        metadata_str = ""
    # Write the output header
    outputs = [
        f"[bold]{position.name}[/]",
        f"┃ [italic dim]{metadata_str}[/]",
    ]
    # Write the motor positions
    for idx, motor in enumerate(position.motors):
        # Figure out some nice tree aesthetics
        is_last_motor = idx == (len(position.motors) - 1)
        box_char = "┗" if is_last_motor else "┣"
        outputs.append(
            f"{box_char}━[purple]{motor.name}[/]: "
            f"{motor.readback}, offset: {motor.offset}"
        )
    rprint("\n".join(outputs))


async def list_motor_positions(after: float | None = None, before: float | None = None):
    """Print a list of previously saved motor positions.

    The name and UID will be printed, along with each motor and it's
    position.

    before
      Only include motor positions recorded before this unix
      timestamp if provided.
    after
      Only include motor positions recorded after this unix
      timestamp if provided.

    """
    # Go through the results and display them
    were_found = False
    async for position in get_motor_positions(before=before, after=after):
        if were_found:
            # Add a blank line
            print("\n")
        else:
            were_found = True
        print_motor_position(position)
    # Some feedback in the case of empty motor positions
    if not were_found:
        rprint(f"[yellow]No motor positions found: {before=}, {after=}[/]")


def get_motor_position(uid: str) -> MotorPosition:
    """Retrieve a previously saved motor position from the database.

    Parameters
    ==========
    uid
      The universal identifier for the Bluesky run with motor position
      info.

    Returns
    =======
    position
      The most recent motor position with data retrieved from the
      database.

    """
    # Filter out query parameters that are ``None``
    client = tiled_client()
    run = client[uid]
    # Feedback for if no matching motor positions are in the database
    position = MotorPosition.load(run)
    return position


async def get_motor_positions(
    before: float | None = None,
    after: float | None = None,
    name: str | None = None,
    case_sensitive: bool = True,
) -> list[MotorPosition]:
    """Get all motor position objects from the catalog.

    Parameters
    ==========
    before
      Only include motor positions recorded before this unix
      timestamp if provided.
    after
      Only include motor positions recorded after this unix
      timestamp if provided.
    name
      A regular expression used to filter motor positions based on
      name.
    case_sensitive
      Whether the regular expression is applied with case-sensitivity.

    Returns
    =======
    positions
      The motor positions matching the requested parameters.

    """
    runs = Catalog(client=tiled_client())
    # Filter only saved motor positions
    runs = await runs.search(Key("plan_name") == "save_motor_position")
    # Filter by timestamp
    if before is not None:
        runs = await runs.search(Key("time") < before)
    if after is not None:
        runs = await runs.search(Key("time") > after)
    # Filter by position name
    if name is not None:
        runs = await runs.search(
            Regex("position_name", name, case_sensitive=case_sensitive)
        )
    # Create the actual motor position objects
    async for uid, run in runs.items():
        try:
            yield await MotorPosition.aload(run)
        except KeyError:
            continue


def recall_motor_position(uid: str):
    """Set motors to their previously saved positions.

    Parameters
    ==========
    uid
      The universal identifier for the the document in the collection.

    """
    # Get the saved position from the database
    position = get_motor_position(uid=uid)
    # Create a move plan to recall the position
    plan_args = []
    for axis in position.motors:
        motor = beamline.registry.find(name=axis.name)
        plan_args.append(motor)
        plan_args.append(axis.readback)
    yield from bps.mv(*plan_args)


async def list_current_motor_positions(*motors, name="Current motor positions"):
    """list and print the current positions of a number of motors

    Parameters
    ==========
    *motors
      The list of motors (or motor names/labels) whose position to
      save.
    name
      A human-readable name for this position (e.g. "sample center")

    """
    # Resolve device names or labels
    motors = [beamline.registry.find(name=m) for m in motors]
    # Build the list of motor positions
    motor_axes = []
    for m in motors:
        payload = dict(name=m.name, readback=await rbv(m))
        # Save the calibration offset for motors
        if hasattr(m, "user_offset"):
            payload["offset"] = await m.user_offset.get_value()
        axis = MotorAxis(**payload)
        motor_axes.append(axis)
    position = MotorPosition(
        name=name, motors=motor_axes, uid=None, savetime=time.time()
    )
    # Display the current motor positions
    print_motor_position(position)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest

from haven.devices.scaler import MultiChannelScaler


@pytest.fixture()
def mcs(sim_registry):
    mcs = MultiChannelScaler("255idcVME:3820:", channels=range(16), name="sis3820")
    return mcs


def test_mcs_signals(mcs):
    # Multi-channel scaler signals
    assert mcs.start_all.source == "ca://255idcVME:3820:StartAll"
    assert mcs.stop_all.source == "ca://255idcVME:3820:StopAll"
    assert mcs.erase_all.source == "ca://255idcVME:3820:EraseAll"
    assert mcs.erase_start.source == "ca://255idcVME:3820:EraseStart"
    assert mcs.preset_time.source == "ca://255idcVME:3820:PresetReal"
    assert mcs.acquiring.source == "ca://255idcVME:3820:Acquiring"
    assert mcs.elapsed_time.source == "ca://255idcVME:3820:ElapsedReal"
    assert mcs.current_channel.source == "ca://255idcVME:3820:CurrentChannel"
    assert mcs.dwell_time.source == "ca://255idcVME:3820:Dwell"
    assert mcs.prescale.source == "ca://255idcVME:3820:Prescale"
    assert mcs.channel_advance_source.source == "ca://255idcVME:3820:ChannelAdvance"
    assert mcs.count_on_start.source == "ca://255idcVME:3820:CountOnStart"
    assert (
        mcs.software_channel_advance.source
        == "ca://255idcVME:3820:SoftwareChannelAdvance"
    )
    assert mcs.channel_1_source.source == "ca://255idcVME:3820:Channel1Source"
    assert mcs.user_led.source == "ca://255idcVME:3820:UserLED"
    assert mcs.mux_output.source == "ca://255idcVME:3820:MUXOutput"
    assert mcs.acquire_mode.source == "ca://255idcVME:3820:AcquireMode"
    assert mcs.input_mode.source == "ca://255idcVME:3820:InputMode"
    assert mcs.input_polarity.source == "ca://255idcVME:3820:InputPolarity"
    assert mcs.output_mode.source == "ca://255idcVME:3820:OutputMode"
    assert mcs.output_polarity.source == "ca://255idcVME:3820:OutputPolarity"
    assert mcs.lne_output_stretcher.source == "ca://255idcVME:3820:LNEStretcherEnable"
    assert mcs.lne_output_polarity.source == "ca://255idcVME:3820:LNEOutputPolarity"
    assert mcs.lne_output_delay.source == "ca://255idcVME:3820:LNEOutputDelay"
    assert mcs.lne_output_width.source == "ca://255idcVME:3820:LNEOutputWidth"
    assert mcs.num_channels_max.source == "ca://255idcVME:3820:MaxChannels"
    assert mcs.num_channels.source == "ca://255idcVME:3820:NuseAll"
    assert mcs.current_channel.source == "ca://255idcVME:3820:CurrentChannel"
    assert mcs.snl_connected.source == "ca://255idcVME:3820:SNL_Connected"
    assert mcs.model.source == "ca://255idcVME:3820:Model"
    assert mcs.firmware.source == "ca://255idcVME:3820:Firmware"


@pytest.mark.asyncio
async def test_mcs_reading(mcs):
    await mcs.connect(mock=True)
    reading = await mcs.read()
    from pprint import pprint

    pprint(list(reading.keys()))
    # Check that the correct readings are included
    assert mcs.elapsed_time.name in reading
    assert mcs.current_channel.name in reading
    assert mcs.mcas[0].spectrum.name in reading
    # The scaler is also read by the MCS by default
    assert mcs.scaler.elapsed_time.name in reading


@pytest.mark.asyncio
async def test_mcs_configuration(mcs):
    await mcs.connect(mock=True)
    config = await mcs.read_configuration()
    # Check that the correct readings are included
    assert mcs.mcas[0].mode.name in config
    assert mcs.preset_time.name in config


@pytest.mark.asyncio
async def test_scaler_reading(mcs):
    await mcs.connect(mock=True)
    scaler = mcs.scaler
    reading = await scaler.read()
    # Check that the correct readings are included
    assert scaler.elapsed_time.name in reading
    assert scaler.channels[0].net_count.name in reading


@pytest.mark.asyncio
async def test_scaler_configuration(mcs):
    await mcs.connect(mock=True)
    scaler = mcs.scaler
    config = await scaler.read_configuration()
    # Check that the correct readings are included
    assert scaler.preset_time.name in config
    assert mcs.preset_time.name not in config


def test_scaler_signals(mcs):
    scaler = mcs.scaler
    # Check individual channel signals
    assert scaler.count.source == "ca://255idcVME:3820:scaler1.CNT"
    assert scaler.count_mode.source == "ca://255idcVME:3820:scaler1.CONT"
    assert scaler.delay.source == "ca://255idcVME:3820:scaler1.DLY"
    assert scaler.auto_count_delay.source == "ca://255idcVME:3820:scaler1.DLY1"
    assert scaler.preset_time.source == "ca://255idcVME:3820:scaler1.TP"
    assert scaler.elapsed_time.source == "ca://255idcVME:3820:scaler1.T"
    assert scaler.auto_count_time.source == "ca://255idcVME:3820:scaler1.TP1"
    assert scaler.clock_frequency.source == "ca://255idcVME:3820:scaler1.FREQ"
    assert (
        scaler.record_dark_current.source
        == "ca://255idcVME:3820:scaler1_offset_start.PROC"
    )
    assert (
        scaler.dark_current_time.source == "ca://255idcVME:3820:scaler1_offset_time.VAL"
    )


def test_mca_signals(mcs):
    mca = mcs.mcas[0]
    assert mca.spectrum.source == "ca://255idcVME:3820:mca1.VAL"
    assert mca.background.source == "ca://255idcVME:3820:mca1.BG"
    assert mca.mode.source == "ca://255idcVME:3820:mca1.MODE"


def test_scaler_channel_signals(mcs):
    # Check individual channel signals
    channel = mcs.scaler.channels[1]
    assert channel.description.source == "ca://255idcVME:3820:scaler1.NM2"
    assert channel.is_gate.source == "ca://255idcVME:3820:scaler1.G2"
    assert channel.preset_count.source == "ca://255idcVME:3820:scaler1.PR2"
    assert channel.raw_count.source == "ca://255idcVME:3820:scaler1.S2"
    assert channel.net_count.source == "ca://255idcVME:3820:scaler1_netA.B"
    assert channel.offset_rate.source == "ca://255idcVME:3820:scaler1_offset0.B"
    # Check that dark current offsets are correct
    channel = mcs.scaler.channels[15]
    assert channel.raw_count.source == "ca://255idcVME:3820:scaler1.S16"
    assert channel.net_count.source == "ca://255idcVME:3820:scaler1_netB.D"
    assert channel.offset_rate.source == "ca://255idcVME:3820:scaler1_offset3.D"



import qtawesome as qta

import haven
from firefly import display


class MotorDisplay(display.FireflyDisplay):
    caqtdm_ui_file = "/APSshare/epics/synApps_6_2_1/support/motor-R7-2-2/motorApp/op/ui/autoconvert/motorx_all.ui"

    def customize_ui(self):
        super().customize_ui()
        self.ui.stop_button.setIcon(qta.icon("fa5s.stop"))

    def ui_filename(self):
        return "motor.ui"

    def launch_caqtdm(self):
        device = haven.registry.find(self.macros()["MOTOR"])
        P, M = device.prefix.split(":")[0:2]
        caqtdm_macros = {
            "P": f"{P}:",
            "M": M,
        }
        super().launch_caqtdm(macros=caqtdm_macros)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



"""Spare definitions for ophyd devices.

This module is deprecated and will eventually be rolled into the
instrument module.

"""

import asyncio
import logging
import re
import time as ttime
import warnings
from typing import Callable, Union

from caproto import CaprotoTimeoutError
from caproto.asyncio.client import Context
from ophyd import Component, Device, K
from ophyd.sim import make_fake_device
from ophyd_async.core import NotConnected

from ._iconfig import load_config

log = logging.getLogger(__name__)

warnings.warn(
    f"Module {__name__} is deprecated and will be removed in a future release.",
    DeprecationWarning,
)


async def connect_devices(
    devices, mock=False, timeout=10.0, labels=None, registry=None
):
    """Ensure that a bunch of ophyd_async devices are connected.

    Returns
    =======
    connected_devices
      Those entries in *devices* that were properly connected.
    """
    aws = (device.connect(mock=mock, timeout=timeout) for device in devices)
    results = await asyncio.gather(*aws, return_exceptions=True)
    # Filter out the disconnected devices
    new_devices = []
    # connection_errors = {}
    exceptions = {}
    for device, result in zip(devices, results):
        if result is None:
            log.debug(f"Successfully connected device {device.name}")
            new_devices.append(device)
        else:
            # Unexpected exception, raise it so it can be handled
            exceptions[device.name] = result
    # Register connected devices with the registry
    if registry is not None:
        for device in new_devices:
            registry.register(device, labels=labels)
    # Raise exceptions if any were present
    if len(exceptions) > 0:
        raise NotConnected(exceptions)
    return new_devices


async def resolve_device_names(ic_defns):
    """Update ion chamber definitions to include the EPICS name field.

    Updates *ic_defns* in place to add the *name* key. Each entry in
    *ic_defns* should have a *desc_pv* key with the PV to be checked
    for the device name.

    Ensures names are safe as Ophyd device names.

    """
    warnings.warn(
        DeprecationWarning(
            "``resolve_device_names`` is deprecated. Perform device name resolution in the device's ``connect()`` method instead."
        )
    )

    async def get_name(pv):
        try:
            response = await pv.read()
        except CaprotoTimeoutError:
            return None
        else:
            # Read PV values over the network
            return response.data.tobytes().strip(b"\x00").decode("latin-1")

    ctx = Context()
    desc_pvs = await ctx.get_pvs(*[defn["desc_pv"] for defn in ic_defns])
    names = await asyncio.gather(*(get_name(pv) for pv in desc_pvs))
    # Add the results back into the defitions
    for name, defn in zip(names, ic_defns):
        defn["name"] = name


def titelize(name):
    """Convert a device name into a human-readable title."""
    title = name.replace("_", " ").title()
    # Replace select phrases that are known to be incorrect
    replacements = {"Kb ": "KB "}
    for orig, new in replacements.items():
        title = title.replace(orig, new)
    return title


async def aload_devices(*coros):
    return await asyncio.gather(*coros)


def make_device(DeviceClass, *args, FakeDeviceClass=None, **kwargs) -> Device:
    """Create device and add it to the registry.

    If the beamline is not connected, i.e. the config file has:

        [beamline]
        is_connected = false

    then the created device will be simulated.

    Parameters
    ==========
    DeviceClass
      The device class to use for making this device.
    FakeDeviceClass
      If the beamline is not connected, use this device instead.

    Returns
    =======
    device
      The newly created and registered camera object.

    """
    # Make a fake device if the beamline is not connected
    config = load_config()
    if config["beamline"]["hardware_is_present"]:
        Cls = DeviceClass
    else:
        # Make fake device
        if FakeDeviceClass is None:
            Cls = make_fake_device(DeviceClass)
        else:
            Cls = FakeDeviceClass
    # Make sure we can connect
    name = kwargs.get("name", "unknown")
    t0 = ttime.monotonic()
    device = Cls(
        *args,
        **kwargs,
    )
    return device


async def await_for_connection(dev, all_signals=False, timeout=3.0):
    """Wait for signals to connect

    Parameters
    ----------
    all_signals : bool, optional
        Wait for all signals to connect (including lazy ones)
    timeout : float or None
        Overall timeout
    """
    signals = [walk.item for walk in dev.walk_signals(include_lazy=all_signals)]

    pending_funcs = {
        dev: getattr(dev, "_required_for_connection", {})
        for name, dev in dev.walk_subdevices(include_lazy=all_signals)
    }
    pending_funcs[dev] = dev._required_for_connection

    t0 = ttime.monotonic()
    # Wait until all the signals have connected
    loop_idx = 0
    connected = False
    while True:
        # Check if the device is ready
        if not connected:
            connected = all(sig.connected for sig in signals)
        if connected and not any(pending_funcs.values()):
            return
        # Since we're not connected, sleep for a short time and try again
        real_timeout = min((0.05, timeout / 10.0))
        tn = ttime.monotonic()
        await asyncio.sleep(real_timeout)
        # Detect other co-routines blocking for too long
        tdelay = ttime.monotonic() - tn
        if tdelay > (real_timeout * 5):
            msg = f"{real_timeout} sec sleep for {dev.name} took "
            msg += f"{tdelay:.4f} sec. "
            msg += "Maybe another co-routine is blocking."
            log.info(msg)
        elapsed_time = ttime.monotonic() - t0
        loop_idx += 1
        if timeout is not None and elapsed_time > timeout and loop_idx > 2:
            break

    def get_name(sig):
        sig_name = f"{dev.name}.{sig.dotted_name}"
        return f"{sig_name} ({sig.pvname})" if hasattr(sig, "pvname") else sig_name

    reasons = []
    unconnected = ", ".join(get_name(sig) for sig in signals if not sig.connected)
    if unconnected:
        reasons.append(f"Failed to connect to all signals: {unconnected}")
    if any(pending_funcs.values()):
        pending = ", ".join(
            description.format(device=dev)
            for dev, funcs in pending_funcs.items()
            for obj, description in funcs.items()
        )
        reasons.append(f"Pending operations: {pending}")
    raise TimeoutError(dev.name + "; ".join(reasons))


class RegexComponent(Component[K]):
    r"""A component with regular expression matching.

    In EPICS, it is not possible to add a field to an existing record,
    e.g. adding a ``.RnXY`` field to go alongside ``mca1.RnNM`` and
    other fields in the MCA record. A common solution is to create a
    new record with an underscore instead of the dot: ``mca1_RnBH``.

    This component include these types of field-like-records as part
    of the ROI device with a ``mca1.Rn`` prefix but performing
    subsitution on the device name using regular expressions. See the
    documentation for ``re.sub`` for full details.

    Example
    =======

    .. code:: python

        class ROI(mca.ROI):
            name = RECpt(EpicsSignal, "NM", lazy=True)
            is_hinted = RECpt(EpicsSignal, "BH",
                              pattern=r"^(.+)\.R(\d+)",
                              repl=r"\1_R\2",
                              lazy=True)

        class MCA(mca.EpicsMCARecord):
            roi0 = Cpt(ROI, ".R0")
            roi1 = Cpt(ROI, ".R1")

        mca = MCA(prefix="mca")
        # *name* has the normal concatination
        assert mca.roi0.name.pvname == "mca.R0NM"
        # *is_hinted* has regex substitution
        assert mca.roi0.is_hinted.pvname == "mca_R0BH"

    """

    def __init__(self, *args, pattern: str, repl: Union[str, Callable], **kwargs):
        """*pattern* and *repl* match their use in ``re.sub``."""
        self.pattern = pattern
        self.repl = repl
        super().__init__(*args, **kwargs)

    def maybe_add_prefix(self, instance, kw, suffix):
        """Parse prefix and suffix with regex suffix if kw is in self.add_prefix.

        Parameters
        ----------
        instance : Device
            The instance from which to extract the prefix to maybe
            append to the suffix.

        kw : str
            The key of associated with the suffix.  If this key is
            self.add_prefix than prepend the prefix to the suffix and
            return, else just return the suffix.

        suffix : str
            The suffix to maybe have something prepended to.

        Returns
        -------
        str

        """
        new_val = super().maybe_add_prefix(instance, kw, suffix)
        try:
            new_val = re.sub(self.pattern, self.repl, new_val)
        except TypeError:
            pass
        return new_val


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from unittest import mock

import pytest
from ophyd import Component as Cpt
from ophyd import Device, EpicsMotor, sim
from ophyd_async.core import Device as AsyncDevice
from ophyd_async.epics.motor import Motor as AsyncMotor

from firefly.component_selector import (
    ComponentSelector,
    DeviceComboBoxModel,
    DeviceTreeModel,
)


class Stage(Device):
    motor2 = Cpt(EpicsMotor, "m2", name="motor2")
    motor3 = Cpt(EpicsMotor, "m3", name="motor3")
    lazy_motor = Cpt(EpicsMotor, "m4", name="motor4", lazy=True)


class StageAsync(AsyncDevice):
    def __init__(self, *args, **kwargs):
        self.motor4 = AsyncMotor("255idcVME:m4")
        self.motor5 = AsyncMotor("255idcVME:m5")
        super().__init__(*args, **kwargs)


@pytest.fixture()
async def motor_registry(sim_registry):
    FakeMotor = sim.make_fake_device(EpicsMotor)
    sim_registry.register(FakeMotor(name="motor1"))
    FakeStage = sim.make_fake_device(Stage)
    sim_registry.register(FakeStage(name="stage"))
    # Add async devices
    async_stage = StageAsync(name="async_stage")
    await async_stage.connect(mock=True)
    sim_registry.register(async_stage)
    return sim_registry


@pytest.fixture
async def selector(motor_registry, qtbot):
    selector_ = ComponentSelector()
    qtbot.addWidget(selector_)
    # Register our new devices
    await selector_.update_devices(motor_registry)
    return selector_


@pytest.mark.asyncio
async def test_selector_adds_devices(selector):
    """Check that the combobox editable options are set based on the allowed detectors."""
    # Check that positioners were added to the combobox model
    num_items = selector.combo_box.count()
    combobox_names = [selector.combo_box.itemText(i) for i in range(num_items)]
    assert "async_stage" in combobox_names
    assert "async_stage.motor4" in combobox_names
    assert "async_stage.motor5" in combobox_names
    assert "motor1" in combobox_names
    assert "stage" in combobox_names
    assert "stage.motor2" in combobox_names
    assert "stage.motor3" in combobox_names
    # Check that devices were added to the tree model
    tree_model = selector.tree_model
    assert tree_model.item(0).text() == "async_stage"


# @pytest.mark.asyncio
# async def test_lazy_signals_omitted(motor_registry, selector):
#     model = DeviceTreeModel()
#     await model.update_devices(motor_registry)
#     stage_idx = 1
#     stage_node = model.item(stage_idx)
#     assert stage_node.text() == "stage"
#     # Make sure the lazy motor isn't included
#     column = 0
#     cpt_names = [stage_node.child(row, column).text() for row in range(stage_node.rowCount())]
#     assert "lazy_motor" not in cpt_names


async def test_selected_component(selector):
    """Does the selector properly report the selected device?"""
    # If no component is selected
    selected = selector.current_component()
    assert selected is None
    # If a component is selected
    selector.combo_box.setCurrentText("async_stage.motor4")
    selected = selector.current_component()
    assert selected is not None
    assert selected.name == "async_stage-motor4"


@pytest.mark.asyncio
async def test_tree_model_adds_device(motor_registry):
    model = DeviceTreeModel()
    await model.add_device(motor_registry["motor1"])
    # Check "Component" column
    assert model.item(0).text() == "motor1"
    assert model.item(0).child(0, column=0).text() == "user_readback"
    # Check "Type" column
    assert model.item(0, column=1).text() == "FakeEpicsMotor"
    assert model.item(0).child(0, column=1).text() == "FakeEpicsSignalRO"


@pytest.mark.asyncio
async def test_combo_box_model_adds_device(motor_registry):
    model = DeviceComboBoxModel()
    await model.add_device(motor_registry["motor1"])
    await model.add_device(motor_registry["stage"])
    # Check that dot-notation is included
    assert model.item(0).text() == "motor1"
    assert model.item(1).text() == "stage"
    assert model.item(2).text() == "stage.motor2"


@pytest.mark.asyncio
async def test_tree_changes_combobox(selector, qtbot):
    """Check that the combobox and tree will update each other."""
    # Select a tree item
    item = selector.tree_model.item(0).child(1, column=0)
    with qtbot.waitSignal(selector.combo_box.currentTextChanged, timeout=1):
        selector.tree_view.selectionModel().currentChanged.emit(
            item.index(), item.index()
        )
    assert selector.combo_box.currentText() == "async_stage.motor5"


@pytest.mark.asyncio
async def test_combobox_changes_tree(selector, qtbot):
    """Check that the combobox and tree will update each other."""
    # Select a combobox item
    item = selector.tree_model.item(0).child(1, column=0)
    selector.combo_box.setCurrentIndex(3)
    with qtbot.waitSignal(
        selector.tree_view.selectionModel().currentChanged, timeout=1
    ):
        selector.combo_box.currentTextChanged.emit("motor1.user_setpoint")


@pytest.mark.asyncio
async def test_model_component_from_index(motor_registry):
    model = DeviceTreeModel()
    await model.add_device(motor_registry["motor1"])
    await model.add_device(motor_registry["stage"])
    # Can we retrieve a root device based on its name item
    item = model.item(0)
    cpt = model.component_from_index(item.index())
    assert cpt.dotted_name == "motor1"
    # Can we retrieve a component based on its name item
    item = model.item(1).child(1, column=0)
    cpt = model.component_from_index(item.index())
    assert cpt.dotted_name == "stage.motor3"
    # Can we retrieve a component based on its type item
    item = model.item(1).child(1, column=1)
    cpt = model.component_from_index(item.index())
    assert cpt.dotted_name == "stage.motor3"


@pytest.mark.asyncio
async def test_model_component_from_dotted_index(motor_registry):
    model = DeviceTreeModel()
    await model.add_device(motor_registry["motor1"])
    await model.add_device(motor_registry["stage"])
    # Can we retrieve a root device based on its dotted name
    item = model.item(0)
    cpt = model.component_from_dotted_name("motor1")
    assert cpt.name_item is item
    # Can we retrieve a component based on its dotted name
    item = model.item(1).child(1, column=0).child(1, column=0)
    cpt = model.component_from_dotted_name("stage.motor3.user_setpoint")
    assert cpt.name_item is item


@pytest.mark.asyncio
async def test_loads_devices_from_registry(selector, motor_registry, qtbot):
    selector.combo_box_model.add_device = mock.AsyncMock()
    await selector.update_devices(motor_registry)
    selector.combo_box_model.add_device.assert_called_with(motor_registry["stage"])



import logging
import subprocess
from collections import OrderedDict
from functools import partial
from pathlib import Path

import pydm
import pyqtgraph as pg
import qtawesome as qta
from ophyd_async.core import NotConnected
from ophydregistry import Registry
from qasync import asyncSlot
from qtpy import QtCore, QtWidgets
from qtpy.QtCore import Signal
from qtpy.QtGui import QIcon, QKeySequence
from qtpy.QtWidgets import QAction, QErrorMessage

from haven import beamline, load_config
from haven.device import titelize
from haven.exceptions import ComponentNotFound, InvalidConfiguration

from .action import Action, ActionsRegistry, WindowAction
from .main_window import FireflyMainWindow, PlanMainWindow
from .queue_client import QueueClient, queueserver_api

generator = type((x for x in []))

__all__ = ["ui_dir", "FireflyController"]


log = logging.getLogger(__name__)


ui_dir = Path(__file__).parent
plans_dir = ui_dir / "plans"


pg.setConfigOption("background", (252, 252, 252))
pg.setConfigOption("foreground", (0, 0, 0))


class FireflyController(QtCore.QObject):
    default_display = None

    # For keeping track of ophyd devices
    registry: Registry = None
    registry_changed = Signal(Registry)

    # For keeping track of the various device and window actions
    actions: ActionsRegistry

    # Signals for running plans on the queueserver
    queue_item_added = Signal(object)

    # Signals responding to queueserver changes
    queue_status_changed = Signal(dict)
    queue_in_use_changed = Signal(bool)  # length > 0, or running
    queue_environment_opened = Signal(bool)  # Opened or closed
    queue_environment_state_changed = Signal(str)  # New state
    queue_manager_state_changed = Signal(str)  # New state
    queue_re_state_changed = Signal(str)  # New state
    queue_empty_changed = Signal(bool)  # Whether queue is empty
    # queue_devices_changed = Signal(dict)  # New list of devices

    def __init__(self, parent=None, display="status", use_main_window=False):
        # Instantiate the parent class
        # (*ui_file* and *use_main_window* let us render the window here instead)
        self.default_display = display
        super().__init__(parent=parent)
        # Initialize needed attributes
        self.actions = ActionsRegistry()
        self.windows = OrderedDict()
        self.queue_re_state_changed.connect(self.enable_queue_controls)
        self.registry = beamline.registry
        # An error message dialog for later use
        self.error_message = QErrorMessage()

    def _setup_window_action(
        self, action_name: str, text: str, slot: QtCore.Slot, shortcut=None, icon=None
    ):
        action = QtWidgets.QAction(self)
        action.setObjectName(action_name)
        action.setText(text)
        action.triggered.connect(slot)
        if shortcut is not None:
            action.setShortcut(QKeySequence(shortcut))
        if icon is not None:
            action.setIcon(qta.icon(icon))
        setattr(self, action_name, action)
        return action

    async def setup_instrument(self, load_instrument=True):
        """Set up the application to use a previously loaded instrument.

        Parameters
        ==========
        load_instrument
          If true, re-read configuration files and create ophyd
          devices. This process is slow.

        Emits
        =====
        registry_changed
          Signal that allows windows to update their widgets for the
          new list of instruments.

        """
        if load_instrument:
            try:
                await beamline.load()
            except NotConnected as exc:
                log.exception(exc)
                msg = (
                    "One or more devices failed to load. See console logs for details."
                )
                self.error_message.showMessage(msg)
            self.registry_changed.emit(beamline.registry)
        # Make actions for launching other windows
        self.setup_window_actions()
        # Actions for controlling the bluesky run engine
        self.setup_queue_actions()
        # Inject menu actions into the various windows
        for action in self.actions.all_actions:
            if hasattr(action, "window_created"):
                action.window_created.connect(self.finalize_new_window)

    def show_default_window(self):
        """Show the first starting window for the application."""
        # Launch the default display
        default_window_action = getattr(self.actions, self.default_display)
        default_window = default_window_action.show_window()
        # Set up the window to show list of PV connections
        pydm.utilities.shortcuts.install_connection_inspector(parent=default_window)

    def setup_window_actions(self):
        """Create QActions for clicking on menu items, shortcuts, etc.

        These actions should be usable by multiple
        windows. Window-specific actions belong with the window.

        """
        # Setup actions for the various categories of devices
        self.actions.motors = self.device_actions(
            device_label="extra_motors",
            display_file=ui_dir / "motor.py",
            device_key="MOTOR",
        )
        self.actions.ion_chambers = self.device_actions(
            device_label="ion_chambers",
            display_file=ui_dir / "ion_chamber.py",
            device_key="IC",
        )
        self.actions.cameras = self.device_actions(
            device_label="cameras",
            display_file=ui_dir / "area_detector_viewer.py",
            device_key="AD",
        )
        self.actions.area_detectors = self.device_actions(
            device_label="area_detectors",
            display_file=ui_dir / "area_detector_viewer.py",
            device_key="AD",
        )
        self.actions.slits = self.device_actions(
            device_label="slits",
            display_file=ui_dir / "slits.py",
            device_key="DEVICE",
            icon=qta.icon("mdi.crop"),
        )
        self.actions.mirrors = self.device_actions(
            device_label="mirrors",
            display_file=ui_dir / "mirror.py",
            device_key="DEVICE",
            icon=qta.icon("msc.mirror"),
        )
        self.actions.mirrors.update(
            self.device_actions(
                device_label="kb_mirrors",
                display_file=ui_dir / "kb_mirrors.py",
                device_key="DEVICE",
                icon=qta.icon("msc.mirror"),
            )
        )
        self.actions.tables = self.device_actions(
            device_label="tables",
            display_file=ui_dir / "table.py",
            device_key="DEVICE",
            icon=qta.icon("mdi.table-furniture"),
        )
        self.actions.robots = self.device_actions(
            device_label="robots",
            display_file=ui_dir / "robot.py",
            device_key="DEVICE",
            WindowClass=PlanMainWindow,
            icon=qta.icon("mdi.robot-industrial"),
        )
        self.actions.xrf_detectors = self.device_actions(
            device_label="xrf_detectors",
            display_file=ui_dir / "xrf_detector.py",
            device_key="DEV",
        )
        self.actions.xray_filter = WindowAction(
            name="show_filters_window_action",
            text="Filters",
            display_file=ui_dir / "filters.py",
            WindowClass=FireflyMainWindow,
            icon=qta.icon("mdi.air-filter"),
        )
        # Action for showing the beamline status window
        self.actions.status = WindowAction(
            name="show_status_window_action",
            text="Beamline Status",
            display_file=ui_dir / "status.py",
            WindowClass=PlanMainWindow,
        )
        self.actions.status.window_created.connect(self.finalize_status_window)
        # Actions for executing plans
        self.actions.plans = {
            "count": WindowAction(
                name="count",
                text="&Count",
                display_file=plans_dir / "count.py",
                shortcut="Ctrl+Shift+C",
                icon=qta.icon("mdi.numeric"),
                WindowClass=PlanMainWindow,
            ),
            "move_motor": WindowAction(
                name="move_motor",
                text="&Move motor",
                display_file=plans_dir / "move_motor_window.py",
                shortcut="Ctrl+Shift+M",
                icon=qta.icon("mdi.rotate-right-variant"),
                WindowClass=PlanMainWindow,
            ),
            "line_scan": WindowAction(
                name="line_scan",
                text="&Line scan",
                display_file=plans_dir / "line_scan.py",
                shortcut="Ctrl+Shift+L",
                icon=qta.icon("mdi.chart-bell-curve"),
                WindowClass=PlanMainWindow,
            ),
            "grid_scan": WindowAction(
                name="grid_scan",
                text="&Grid scan",
                shortcut="Ctrl+Shift+G",
                display_file=plans_dir / "grid_scan.py",
                icon=qta.icon("mdi.grid"),
                WindowClass=PlanMainWindow,
            ),
            "xafs_scan": WindowAction(
                name="xafs_scan",
                text="&XAFS scan",
                display_file=plans_dir / "xafs_scan.py",
                shortcut="Ctrl+Shift+X",
                icon=qta.icon("mdi.chart-bell-curve-cumulative"),
                WindowClass=PlanMainWindow,
            ),
        }

        # Action for showing the run browser window
        self.actions.run_browser = WindowAction(
            name="show_run_browser_action",
            text="Browse Runs",
            display_file=ui_dir / "run_browser.py",
            shortcut="Ctrl+Shift+B",
            icon=qta.icon("mdi.book-open-variant"),
            WindowClass=FireflyMainWindow,
        )
        # Action for showing the beamline scheduling window
        self.actions.bss = WindowAction(
            name="show_bss_window_action",
            text="Scheduling (&BSS)",
            display_file=ui_dir / "bss.py",
            WindowClass=FireflyMainWindow,
            icon=qta.icon("fa5s.calendar"),
        )
        # Action for shoing the IOC start/restart/stop window
        self.actions.iocs = WindowAction(
            name="show_iocs_window_action",
            text="&IOCs",
            display_file=ui_dir / "iocs.py",
            WindowClass=FireflyMainWindow,
        )
        # Launch ion chamber voltmeters window
        self.actions.voltmeter = WindowAction(
            name="show_voltmeters_window_action",
            text="&Voltmeters",
            display_file=ui_dir / "voltmeters.py",
            WindowClass=FireflyMainWindow,
            shortcut="Ctrl+V",
            icon=qta.icon("ph.faders-horizontal"),
        )
        self.actions.voltmeter.window_created.connect(self.finalize_voltmeter_window)
        # Launch log window
        self.actions.log = WindowAction(
            name="show_logs_window_action",
            text="Logs",
            display_file=ui_dir / "log_viewer.ui",
            WindowClass=FireflyMainWindow,
            icon=qta.icon("mdi.view-list-outline"),
        )
        # Launch energy window
        self.actions.energy = WindowAction(
            name="show_energy_window_action",
            text="Energy",
            display_file=ui_dir / "energy.py",
            WindowClass=PlanMainWindow,
            shortcut="Ctrl+E",
            icon=qta.icon("mdi.sine-wave"),
        )

    def update_queue_controls(self, status):
        print(status)

    @asyncSlot(QAction)
    async def finalize_new_window(self, action):
        """Slot for providing new windows for after a new window is created."""
        action.window.setup_menu_actions(actions=self.actions)
        self.queue_status_changed.connect(action.window.update_queue_status)
        self.queue_status_changed.connect(action.window.update_queue_controls)
        if getattr(self, "_queue_client", None) is not None:
            self._queue_client.check_queue_status(force=True)
        action.display.queue_item_submitted.connect(self.add_queue_item)
        # Send the current devices to the window
        await action.window.update_devices(self.registry)

    def finalize_status_window(self, action):
        """Connect up signals that are specific to the voltmeters window."""
        display = action.display
        display.ui.bss_modify_button.clicked.connect(self.actions.bss.trigger)
        # display.details_window_requested.connect

    def finalize_voltmeter_window(self, action):
        """Connect up signals that are specific to the voltmeters window."""

        def launch_ion_chamber_window(ic_name):
            action = self.actions.ion_chambers[ic_name]
            action.trigger()

        display = action.window.display_widget()
        display.details_window_requested.connect(launch_ion_chamber_window)

    def launch_queuemonitor(self):
        config = load_config()["queueserver"]
        zmq_info_addr = f"tcp://{config['info_host']}:{config['info_port']}"
        zmq_ctrl_addr = f"tcp://{config['control_host']}:{config['control_port']}"
        cmds = [
            "queue-monitor",
            "--zmq-control-addr",
            zmq_ctrl_addr,
            "--zmq-info-addr",
            zmq_info_addr,
        ]
        subprocess.Popen(cmds)

    def setup_queue_actions(self):
        """Create QAction objects for controlling the bluesky queueserver."""
        # Internal actions for interacting with the run engine
        self.check_queue_status_action = QtWidgets.QAction(self)
        # Action for launch queue-monitor
        self.actions.queue_monitor = Action(
            name="launch_queuemonitor_action",
            text="Queue Monitor",
            shortcut="Ctrl+Q",
        )
        self.actions.queue_monitor.triggered.connect(self.launch_queuemonitor)
        # Navbar actions for controlling the run engine
        self.actions.queue_controls = {
            "pause": Action(
                name="pause_runengine_action",
                text="Pause",
                shortcut="Ctrl+D",
                icon=qta.icon("fa5s.stopwatch"),
                tooltip="Pause the current plan at the next checkpoint.",
            ),
            "pause_now": Action(
                name="pause_runengine_now_action",
                text="Pause now",
                shortcut="Ctrl+C",
                icon=qta.icon("fa5s.pause"),
                tooltip="Pause the run engine now.",
            ),
            "resume": Action(
                name="resume_runengine_action",
                text="Resume",
                icon=qta.icon("fa5s.play"),
                tooltip="Resume a paused run engine at the last checkpoint.",
            ),
            "stop_runengine": Action(
                name="stop_runengine_action",
                text="Success",
                icon=qta.icon("fa5s.check"),
                tooltip="End the current plan, marking as successful.",
                checkable=True,
            ),
            "abort": Action(
                name="abort_runengine_action",
                text="Abort",
                icon=qta.icon("fa5s.times"),
                tooltip="End the current plan, marking as failure.",
            ),
            "start": Action(
                name="start_queue_action",
                text="Start",
                icon=qta.icon("fa5s.play"),
                tooltip="Start the queue",
            ),
            "halt": Action(
                name="halt_runengine_action",
                text="Halt",
                icon=qta.icon("fa5s.ban"),
                tooltip="End the current plan immediately, do not clean up.",
            ),
            "stop_queue": Action(
                name="stop_queue_action",
                text="Stop Queue",
                icon=qta.icon("fa5s.stop"),
                tooltip="Instruct the queue to stop after the current item is done.",
                checkable=True,
            ),
        }
        # Actions that control how the queue operates
        self.actions.queue_settings = {
            "autostart": Action(
                name="queue_autostart_action",
                text="&Autoplay",
                tooltip="If enabled, the queue will start when items are added.",
                checkable=True,
            ),
            "open_environment": Action(
                name="queue_open_environment_action",
                text="&Open Environment",
                tooltip="If open (checked), the queue server is able to run plans.",
                checkable=True,
            ),
        }

    def device_actions(
        self,
        device_label: str,
        display_file: str = None,
        device_key: str = "DEVICE",
        WindowClass: type = FireflyMainWindow,
        icon: QIcon = None,
    ):
        """Generic routine to be called for individual classes of devices.

        Sets up a window action for each instance of this device class
        (specified by *device_label*).

        For example, to set up device window actions for all a Tardis
        (Ophyd devices with the "tardis_ship" label), call:

        .. code:: python

            app.device_actions(device_label="tardis_ship")

        Parameters
        ==========
        device_label
          The Ophyd label by which to find the devices.
        display_file
          A path object pointing to the .ui or .py file to use for
          this device.
        device_key
          A key to use for the device name in the macros
          dictionary. If *device_key* is "DEVICE" (default), then the
          macros will be {"DEVICE": device.name}. Has no effect if
          *window_slot* is used.
        WindowClass
          The type of window to create around this device display.
        icon
          A QIcon that will be added to the action.

        """
        # We need a UI file, unless a custom window_slot is given
        # Get needed devices from the device registry
        try:
            devices = sorted(
                self.registry.findall(label=device_label), key=lambda x: x.name
            )
        except ComponentNotFound:
            log.warning(f"No {device_label} found, menu will be empty.")
            devices = []
        # Create menu actions for each device
        actions = {
            device.name: WindowAction(
                name=f"show_{device.name}_action",
                text=titelize(device.name),
                display_file=display_file,
                icon=icon,
                WindowClass=WindowClass,
                macros={device_key: device.name},
            )
            for device in devices
        }
        return actions

    def start_queue_client(self):
        try:
            self._queue_client.start()
        except Exception as exc:
            log.error(f"Could not start queue client: {exc}")

    def prepare_queue_client(self, client=None, api=None):
        """Set up the QueueClient object that talks to the queue server.

        Parameters
        ==========
        api
          queueserver API. Used for testing.

        """
        # Load the API for controlling the queueserver.
        if api is None:
            try:
                api = queueserver_api()
            except InvalidConfiguration:
                log.error(
                    "Could not load queueserver API "
                    "configuration from iconfig.toml file."
                )
                return
        # Create the client object
        if client is None:
            client = QueueClient(api=api)
        # self.queue_open_environment_action.triggered.connect(client.open_environment)
        self.actions.queue_settings["open_environment"].triggered.connect(
            client.open_environment
        )
        self._queue_client = client
        # Connect actions to slots for controlling the queueserver
        self.actions.queue_controls["pause"].triggered.connect(
            partial(client.request_pause, defer=True)
        )
        self.actions.queue_controls["pause_now"].triggered.connect(
            partial(client.request_pause, defer=False)
        )
        self.actions.queue_controls["start"].triggered.connect(client.start_queue)
        self.actions.queue_controls["resume"].triggered.connect(client.resume_runengine)
        self.actions.queue_controls["stop_runengine"].triggered.connect(
            client.stop_runengine
        )
        self.actions.queue_controls["halt"].triggered.connect(client.halt_runengine)
        self.actions.queue_controls["abort"].triggered.connect(client.abort_runengine)
        self.actions.queue_controls["stop_queue"].triggered.connect(client.stop_queue)
        self.check_queue_status_action.triggered.connect(
            partial(client.check_queue_status, True)
        )
        # Connect signals/slots for queueserver state changes
        client.status_changed.connect(self.queue_status_changed)
        client.in_use_changed.connect(self.queue_in_use_changed)
        client.autostart_changed.connect(
            self.actions.queue_settings["autostart"].setChecked
        )
        client.environment_opened.connect(self.queue_environment_opened)
        client.environment_opened.connect(self.queue_environment_opened)
        self.queue_environment_opened.connect(self.set_open_environment_action_state)
        client.environment_state_changed.connect(self.queue_environment_state_changed)
        client.manager_state_changed.connect(self.queue_manager_state_changed)
        client.re_state_changed.connect(self.queue_re_state_changed)
        client.queue_stop_changed.connect(
            self.actions.queue_controls["stop_queue"].setChecked
        )
        client.devices_changed.connect(self.update_devices_allowed)
        self.actions.queue_settings["autostart"].toggled.connect(
            self.check_queue_status_action.trigger
        )
        self.actions.queue_settings["autostart"].toggled.connect(
            client.toggle_autostart
        )
        return client

    def start(self):
        """Start the background clients."""
        # Show the UI stuffs
        self.prepare_queue_client()
        self.start_queue_client()

    def update_devices_allowed(self, devices):
        pass

    def enable_queue_controls(self, re_state):
        """Enable/disable the navbar buttons that control the queue.

        Most buttons are only relevant when the run engine is in
        certain states. For exmple, you can't click Play if the run
        engine is already running.

        """
        queue_actions = self.actions.queue_controls
        # Decide which signals to enable
        unknown_re_state = re_state is None or re_state.strip() == ""
        if unknown_re_state:
            # Unknown state, no button should work
            enabled_signals = []
        elif re_state == "idle":
            enabled_signals = [queue_actions["start"]]
        elif re_state == "paused":
            enabled_signals = [
                queue_actions["stop_runengine"],
                queue_actions["resume"],
                queue_actions["halt"],
                queue_actions["abort"],
            ]
        elif re_state == "running":
            enabled_signals = [
                queue_actions["pause"],
                queue_actions["pause_now"],
                queue_actions["stop_queue"],
            ]
        elif re_state in ["stopping", "aborting"]:
            enabled_signals = []
        else:
            enabled_signals = []
            raise ValueError(f"Unknown run engine state: {re_state}")
        # Enable/disable the relevant signals
        for action in queue_actions.values():
            action.setEnabled(action in enabled_signals)

    @asyncSlot(object)
    async def add_queue_item(self, item):
        log.debug(f"Application received item to add to queue: {item}")
        if getattr(self, "_queue_client", None) is not None:
            await self._queue_client.add_queue_item(item)

    @QtCore.Slot()
    def show_sample_viewer_window(self):
        return self.show_window(
            FireflyMainWindow, ui_dir / "sample_viewer.ui", name="sample_viewer"
        )

    @QtCore.Slot(bool)
    def set_open_environment_action_state(self, is_open: bool):
        """Update the readback value for opening the queueserver environment."""
        action = self.actions.queue_settings["open_environment"]
        if action is not None:
            action.blockSignals(True)
            action.setChecked(is_open)
            action.blockSignals(False)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import asyncio
from unittest.mock import MagicMock

import numpy as np
import pytest
from pyqtgraph import ImageItem, ImageView, PlotItem, PlotWidget
from qtpy.QtWidgets import QFileDialog

from firefly.run_browser import RunBrowserDisplay
from firefly.run_client import DatabaseWorker


@pytest.fixture()
async def display(qtbot, catalog, mocker):
    mocker.patch(
        "firefly.run_browser.ExportDialog.exec_", return_value=QFileDialog.Accepted
    )
    mocker.patch(
        "firefly.run_browser.ExportDialog.selectedFiles",
        return_value=["/net/s255data/export/test_file.nx"],
    )
    mocker.patch("firefly.run_client.DatabaseWorker.export_runs")
    display = RunBrowserDisplay(root_node=catalog)
    qtbot.addWidget(display)
    display.clear_filters()
    # Wait for the initial database load to process
    await display._running_db_tasks["init_load_runs"]
    await display._running_db_tasks["update_combobox_items"]
    return display


@pytest.mark.asyncio
async def test_db_task(display):
    async def test_coro():
        return 15

    result = await display.db_task(test_coro())
    assert result == 15


@pytest.mark.asyncio
async def test_db_task_interruption(display):
    async def test_coro(sleep_time):
        await asyncio.sleep(sleep_time)
        return sleep_time

    # Create an existing task that will be cancelled
    task_1 = display.db_task(test_coro(1.0), name="testing")
    # Now execute another task
    result = await display.db_task(test_coro(0.01), name="testing")
    assert result == 0.01
    # Check that the first one was cancelled
    with pytest.raises(asyncio.exceptions.CancelledError):
        await task_1
    assert task_1.done()
    assert task_1.cancelled()


def test_load_runs(display):
    assert display.runs_model.rowCount() > 0
    assert display.ui.runs_total_label.text() == str(display.runs_model.rowCount())


@pytest.mark.asyncio
async def test_update_selected_runs(display):
    # Change the proposal item
    item = display.runs_model.item(0, 1)
    assert item is not None
    display.ui.run_tableview.selectRow(0)
    # Update the runs
    await display.update_selected_runs()
    # Check that the runs were saved
    assert len(display.db.selected_runs) > 0


@pytest.mark.asyncio
async def test_metadata(display):
    # Change the proposal item
    display.ui.run_tableview.selectRow(0)
    await display.update_selected_runs()
    # Check that the metadata was set properly in the Metadata tab
    text = display.ui.metadata_textedit.document().toPlainText()
    assert "xafs_scan" in text


@pytest.mark.asyncio
async def test_1d_plot_signals(catalog, display):
    # Check that the 1D plot was created
    plot_widget = display.ui.plot_1d_view
    plot_item = display.plot_1d_item
    assert isinstance(plot_widget, PlotWidget)
    assert isinstance(plot_item, PlotItem)
    # Update the list of runs and see if the controls get updated
    display.ui.run_tableview.selectColumn(0)
    await display.update_selected_runs()
    # Check signals in checkboxes
    for combobox in [
        display.ui.signal_y_combobox,
        display.ui.signal_r_combobox,
        display.ui.signal_x_combobox,
    ]:
        assert (
            combobox.findText("energy_energy") > -1
        ), f"energy_energy signal not in {combobox.objectName()}."


# Warns: Task was destroyed but it is pending!
@pytest.mark.asyncio
async def test_1d_plot_signal_memory(catalog, display):
    """Do we remember the signals that were previously selected."""
    # Check that the 1D plot was created
    plot_widget = display.ui.plot_1d_view
    plot_item = display.plot_1d_item
    assert isinstance(plot_widget, PlotWidget)
    assert isinstance(plot_item, PlotItem)
    # Update the list of runs and see if the controls get updated
    display.ui.run_tableview.selectRow(1)
    await display.update_selected_runs()
    # Check signals in comboboxes
    cb = display.ui.signal_y_combobox
    assert cb.currentText() == "energy_energy"
    cb.setCurrentIndex(1)
    assert cb.currentText() == "energy_id_energy_readback"
    # Update the combobox signals and make sure the text didn't change
    await display.update_1d_signals()
    assert cb.currentText() == "energy_id_energy_readback"


# Warns: Task was destroyed but it is pending!
@pytest.mark.asyncio
async def test_1d_hinted_signals(catalog, display):
    display.ui.plot_1d_hints_checkbox.setChecked(True)
    # Check that the 1D plot was created
    plot_widget = display.ui.plot_1d_view
    plot_item = display.plot_1d_item
    assert isinstance(plot_widget, PlotWidget)
    assert isinstance(plot_item, PlotItem)
    # Update the list of runs and see if the controls get updated
    display.db.selected_runs = [run async for run in catalog.values()]
    await display.update_1d_signals()
    return
    # Check signals in checkboxes
    combobox = display.ui.signal_x_combobox
    assert (
        combobox.findText("energy_energy") > -1
    ), f"hinted signal not in {combobox.objectName()}."
    assert (
        combobox.findText("It_net_counts") == -1
    ), f"unhinted signal found in {combobox.objectName()}."


@pytest.mark.asyncio
async def test_update_1d_plot(catalog, display):
    # Set up some fake data
    run = [run async for run in catalog.values()][0]
    display.db.selected_runs = [run]
    await display.update_1d_signals()
    run_data = await run.to_dataframe()
    expected_xdata = run_data.energy_energy
    expected_ydata = np.log(run_data.I0_net_counts / run_data.It_net_counts)
    expected_ydata = np.gradient(expected_ydata, expected_xdata)
    # Set the controls to describe the data we want to test
    x_combobox = display.ui.signal_x_combobox
    x_combobox.addItem("energy_energy")
    x_combobox.setCurrentText("energy_energy")
    y_combobox = display.ui.signal_y_combobox
    y_combobox.addItem("It_net_counts")
    y_combobox.setCurrentText("It_net_counts")
    r_combobox = display.ui.signal_r_combobox
    r_combobox.addItem("I0_net_counts")
    r_combobox.setCurrentText("I0_net_counts")
    display.ui.signal_r_checkbox.setChecked(True)
    display.ui.logarithm_checkbox.setChecked(True)
    display.ui.invert_checkbox.setChecked(True)
    display.ui.gradient_checkbox.setChecked(True)
    # Update the plots
    await display.update_1d_plot()
    # Check that the data were added
    data_item = display.plot_1d_item.listDataItems()[0]
    xdata, ydata = data_item.getData()
    np.testing.assert_almost_equal(xdata, expected_xdata)
    np.testing.assert_almost_equal(ydata, expected_ydata)


# Warns: Task was destroyed but it is pending!
@pytest.mark.asyncio
async def test_2d_plot_signals(catalog, display):
    # Check that the 1D plot was created
    plot_widget = display.ui.plot_2d_view
    plot_item = display.plot_2d_item
    assert isinstance(plot_widget, ImageView)
    assert isinstance(plot_item, ImageItem)
    # Update the list of runs and see if the controls get updated
    display.db.selected_runs = [await catalog["85573831-f4b4-4f64-b613-a6007bf03a8d"]]
    await display.update_2d_signals()
    # Check signals in checkboxes
    combobox = display.ui.signal_value_combobox
    assert combobox.findText("It_net_counts") > -1


@pytest.mark.asyncio
async def test_update_2d_plot(catalog, display):
    display.plot_2d_item.setRect = MagicMock()
    # Load test data
    run = await catalog["85573831-f4b4-4f64-b613-a6007bf03a8d"]
    display.db.selected_runs = [run]
    await display.update_1d_signals()
    # Set the controls to describe the data we want to test
    val_combobox = display.ui.signal_value_combobox
    val_combobox.addItem("It_net_counts")
    val_combobox.setCurrentText("It_net_counts")
    display.ui.logarithm_checkbox_2d.setChecked(True)
    display.ui.invert_checkbox_2d.setChecked(True)
    display.ui.gradient_checkbox_2d.setChecked(True)
    # Update the plots
    await display.update_2d_plot()
    # Determine what the image data should look like
    expected_data = await run["It_net_counts"]
    expected_data = expected_data.reshape((5, 21)).T
    # Check that the data were added
    image = display.plot_2d_item.image
    np.testing.assert_almost_equal(image, expected_data)
    # Check that the axes were formatted correctly
    axes = display.plot_2d_view.view.axes
    xaxis = axes["bottom"]["item"]
    yaxis = axes["left"]["item"]
    assert xaxis.labelText == "aerotech_horiz"
    assert yaxis.labelText == "aerotech_vert"
    display.plot_2d_item.setRect.assert_called_with(-100, -80, 200, 160)


@pytest.mark.asyncio
async def test_update_multi_plot(catalog, display):
    run = await catalog["7d1daf1d-60c7-4aa7-a668-d1cd97e5335f"]
    expected_xdata = await run["energy_energy"]
    expected_ydata = np.log(await run["I0_net_counts"] / await run["It_net_counts"])
    expected_ydata = np.gradient(expected_ydata, expected_xdata)
    # Configure signals
    display.ui.multi_signal_x_combobox.addItem("energy_energy")
    display.ui.multi_signal_x_combobox.setCurrentText("energy_energy")
    display.multi_y_signals = ["energy_energy"]
    display.db.selected_runs = [run]
    # Update the plots
    await display.update_multi_plot()
    # Check that the data were added
    # data_item = display._multiplot_items[0].listDataItems()[0]
    # xdata, ydata = data_item.getData()
    # np.testing.assert_almost_equal(xdata, expected_xdata)
    # np.testing.assert_almost_equal(ydata, expected_ydata)


@pytest.mark.asyncio
async def test_filter_runs(catalog):
    worker = DatabaseWorker(catalog=catalog)
    runs = await worker.load_all_runs(filters={"plan": "xafs_scan"})
    # Check that the runs were filtered
    assert len(runs) == 1


@pytest.mark.asyncio
async def test_distinct_fields(catalog, display):
    worker = DatabaseWorker(catalog=catalog)
    distinct_fields = await worker.load_distinct_fields()
    # Check that the dictionary has the right structure
    for key in ["sample_name"]:
        assert key in distinct_fields.keys()


def test_busy_hints_run_widgets(display):
    """Check that the display widgets get disabled during DB hits."""
    with display.busy_hints(run_widgets=True, run_table=False):
        # Are widgets disabled in the context block?
        assert not display.ui.detail_tabwidget.isEnabled()
    # Are widgets re-enabled outside the context block?
    assert display.ui.detail_tabwidget.isEnabled()


def test_busy_hints_run_table(display):
    """Check that the all_runs table view gets disabled during DB hits."""
    with display.busy_hints(run_table=True, run_widgets=False):
        # Are widgets disabled in the context block?
        assert not display.ui.run_tableview.isEnabled()
    # Are widgets re-enabled outside the context block?
    assert display.ui.run_tableview.isEnabled()


def test_busy_hints_filters(display):
    """Check that the all_runs table view gets disabled during DB hits."""
    with display.busy_hints(run_table=False, run_widgets=False, filter_widgets=True):
        # Are widgets disabled in the context block?
        assert not display.ui.filters_widget.isEnabled()
    # Are widgets re-enabled outside the context block?
    assert display.ui.filters_widget.isEnabled()


def test_busy_hints_status(display, mocker):
    """Check that any busy_hints displays the message "Loading…"."""
    spy = mocker.spy(display, "show_message")
    with display.busy_hints(run_table=True, run_widgets=False):
        # Are widgets disabled in the context block?
        assert not display.ui.run_tableview.isEnabled()
        assert spy.call_count == 1
    # Are widgets re-enabled outside the context block?
    assert spy.call_count == 2
    assert display.ui.run_tableview.isEnabled()


def test_busy_hints_multiple(display):
    """Check that multiple busy hints can co-exist."""
    # Next the busy_hints context to mimic multiple async calls
    with display.busy_hints(run_widgets=True):
        # Are widgets disabled in the outer block?
        assert not display.ui.detail_tabwidget.isEnabled()
        with display.busy_hints(run_widgets=True):
            # Are widgets disabled in the inner block?
            assert not display.ui.detail_tabwidget.isEnabled()
        # Are widgets still disabled in the outer block?
        assert not display.ui.detail_tabwidget.isEnabled()
    # Are widgets re-enabled outside the context block?
    assert display.ui.detail_tabwidget.isEnabled()


@pytest.mark.asyncio
async def test_update_combobox_items(display):
    """Check that the comboboxes get the distinct filter fields."""
    assert display.ui.filter_plan_combobox.count() > 0


@pytest.mark.asyncio
async def test_export_button_enabled(catalog, display):
    assert not display.export_button.isEnabled()
    # Update the list with 1 run and see if the control gets enabled
    display.selected_runs = [run async for run in catalog.values()]
    display.selected_runs = display.selected_runs[:1]
    display.update_export_button()
    assert display.export_button.isEnabled()
    # Update the list with multiple runs and see if the control gets disabled
    display.selected_runs = [run async for run in catalog.values()]
    display.update_export_button()
    assert not display.export_button.isEnabled()


@pytest.mark.asyncio
async def test_export_button_clicked(catalog, display, mocker, qtbot):
    # Set up a run to be tested against
    run = MagicMock()
    run.formats.return_value = [
        "application/json",
        "application/x-hdf5",
        "application/x-nexus",
    ]
    display.selected_runs = [run]
    display.update_export_button()
    # Clicking the button should open a file dialog
    await display.export_runs()
    assert display.export_dialog.exec_.called
    assert display.export_dialog.selectedFiles.called
    # Check that file filter names are set correctly
    # (assumes application/json is available on every machine)
    assert "JSON document (*.json)" in display.export_dialog.nameFilters()
    # Check that the file was saved
    assert display.db.export_runs.called
    files = display.export_dialog.selectedFiles.return_value
    assert display.db.export_runs.call_args.args == (files,)
    assert display.db.export_runs.call_args.kwargs["formats"] == ["application/json"]


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import time
from unittest.mock import MagicMock

import pytest
from ophyd.sim import make_fake_device
from pydm import PyDMChannel
from pydm.data_plugins import plugin_for_address
from pydm.main_window import PyDMMainWindow
from pydm.widgets import PyDMLineEdit
from qtpy import QtCore
from typhos.plugins.core import SignalPlugin

from haven import HavenMotor


class DummyObject(QtCore.QObject):
    signal = QtCore.Signal()


@pytest.fixture()
def sim_motor(sim_registry):
    sim_registry.use_typhos = True
    FakeMotor = make_fake_device(HavenMotor)
    motor = FakeMotor("255idVME:m1", name="motor")
    motor.user_setpoint.sim_set_limits((0, 1000))
    sim_registry.register(motor)
    return motor


@pytest.fixture()
def ophyd_channel(sim_motor):
    channel = PyDMChannel(address="haven://motor.user_setpoint")
    return channel


@pytest.fixture()
def ophyd_connection(sim_motor, ophyd_channel, pydm_ophyd_plugin):
    channel = ophyd_channel
    pydm_ophyd_plugin.add_connection(channel)
    connection = pydm_ophyd_plugin.connections["motor.user_setpoint"]
    yield connection
    pydm_ophyd_plugin.remove_connection(channel)


def test_ophyd_pydm_ophyd_plugin(pydm_ophyd_plugin):
    plugin = plugin_for_address("haven://sim_detector")
    assert isinstance(plugin, SignalPlugin)


@pytest.mark.skip(reason="Moved to typhos, test should be removed in the future")
def test_new_value(sim_motor, ophyd_connection, qtbot):
    with qtbot.waitSignal(ophyd_connection.new_value_signal, timeout=1000) as blocker:
        sim_motor.set(45.0)
    assert blocker.args[0] == 45.0


# def test_set_value(sim_motor, ophyd_connection, ophyd_channel):
#     ophyd_connection.set_value(87.0).wait()
#     assert sim_motor.user_setpoint.get(use_monitor=False) == 87.0


def test_missing_device(sim_motor, pydm_ophyd_plugin):
    """See if the connection responds properly if the device is not there."""
    connection_slot = MagicMock()
    channel = PyDMChannel(
        address="haven://motor.nonsense_parts", connection_slot=connection_slot
    )
    pydm_ophyd_plugin.add_connection(channel)


# def test_update_ctrl_vals(sim_motor, ophyd_connection, qtbot):
#     conn = ophyd_connection
#     conn._ctrl_vars = {}
#     # Check if the connection state is updated
#     with qtbot.waitSignal(conn.connection_state_signal) as blocker:
#         conn.update_ctrl_vars(connected=True)
#     assert blocker.args[0] is True
#     # Make sure it isn't emitted a second time if it doesn't change
#     conn._ctrl_vars = {"connected": True}
#     conn.connection_state_signal = MagicMock()
#     conn.update_ctrl_vars(connected=True)
#     assert not conn.connection_state_signal.emit.called
#     # Check other metadata signals
#     conn._ctrl_vars = {}
#     with qtbot.waitSignal(conn.new_severity_signal) as blocker:
#         conn.update_ctrl_vars(severity=2)
#     assert blocker.args[0] == 2

#     conn._ctrl_vars = {}
#     with qtbot.waitSignal(conn.write_access_signal) as blocker:
#         conn.update_ctrl_vars(write_access=True)
#     assert blocker.args[0] is True

#     conn._ctrl_vars = {}
#     with qtbot.waitSignal(conn.enum_strings_signal) as blocker:
#         conn.update_ctrl_vars(enum_strs=("Option 1", "Option 2"))
#     assert blocker.args[0] == ("Option 1", "Option 2")

#     conn._ctrl_vars = {}
#     with qtbot.waitSignal(conn.unit_signal) as blocker:
#         conn.update_ctrl_vars(units="µm")
#     assert blocker.args[0] == "µm"

#     conn._ctrl_vars = {}
#     with qtbot.waitSignal(conn.prec_signal) as blocker:
#         conn.update_ctrl_vars(precision=5)
#     assert blocker.args[0] == 5

#     conn._ctrl_vars = {}
#     with qtbot.waitSignal(conn.lower_ctrl_limit_signal) as blocker:
#         conn.update_ctrl_vars(lower_ctrl_limit=-200)
#     assert blocker.args[0] == -200

#     conn._ctrl_vars = {}
#     with qtbot.waitSignal(conn.upper_ctrl_limit_signal) as blocker:
#         conn.update_ctrl_vars(upper_ctrl_limit=850.3)
#     assert blocker.args[0] == 850.3

#     conn._ctrl_vars = {}
#     with qtbot.waitSignal(conn.timestamp_signal) as blocker:
#         conn.update_ctrl_vars(timestamp=1693014035.3913143)
#     assert blocker.args[0] == 1693014035.3913143
#     {
#         "connected": True,
#         "read_access": True,
#         "write_access": True,
#         "timestamp": 1693014035.3913143,
#         "status": None,
#         "severity": None,
#         "precision": None,
#         "lower_ctrl_limit": None,
#         "upper_ctrl_limit": None,
#         "units": None,
#         "enum_strs": None,
#         "setpoint_status": None,
#         "setpoint_severity": None,
#         "setpoint_precision": None,
#         "setpoint_timestamp": None,
#         "sub_type": "meta",
#     }


def test_widget_signals(sim_motor, qapp, qtbot):
    """Does this work with a real widget in a real window."""
    sim_motor.user_setpoint.set(5.15)
    sim_motor.user_setpoint._metadata["precision"] = 3
    window = PyDMMainWindow()
    widget = PyDMLineEdit(parent=window, init_channel="haven://motor.user_setpoint")
    qapp.processEvents()
    time.sleep(0.05)
    assert widget.text() == "5.150"
    # Now check that we can set the widget text and have it update the ophyd device
    widget.send_value_signal[float].emit(4.9)
    qapp.processEvents()
    time.sleep(0.05)
    assert sim_motor.user_setpoint.get(use_monitor=False) == 4.9


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from typing import Mapping

from apsbss.apsbss_ophyd import EpicsBssDevice, EpicsEsafDevice, EpicsProposalDevice
from ophyd import Component as Cpt
from ophyd import Device
from ophyd import DynamicDeviceComponent as DCpt
from ophyd import EpicsSignal, EpicsSignalRO


class BSSEsaf(EpicsEsafDevice):
    user_PIs = Cpt(EpicsSignal, "userPIs", string=True)
    start_timestamp = Cpt(EpicsSignal, "startTimestamp")
    end_timestamp = Cpt(EpicsSignal, "endTimestamp")


class BSSProposal(EpicsProposalDevice):
    user_PIs = Cpt(EpicsSignal, "userPIs", string=True)
    mail_in_flag = Cpt(EpicsSignal, "mailInFlag", string=False)
    proprietary_flag = Cpt(EpicsSignal, "proprietaryFlag", string=False)
    start_timestamp = Cpt(EpicsSignal, "startTimestamp")
    end_timestamp = Cpt(EpicsSignal, "endTimestamp")


class BSS(EpicsBssDevice):
    esaf = Cpt(BSSEsaf, "esaf:")
    proposal = Cpt(BSSProposal, "proposal:")


class LocalStorage(Device):
    file_system = Cpt(EpicsSignal, "file_system", kind="config")
    sub_directory = Cpt(EpicsSignal, "sub_directory", kind="config")
    full_path = Cpt(EpicsSignalRO, "full_path", kind="config")
    exists = Cpt(EpicsSignalRO, "exists", kind="config")

    create = Cpt(EpicsSignal, "create", kind="omitted")


class IOCManager(Device):
    startable = Cpt(EpicsSignal, "startable", kind="config")
    stoppable = Cpt(EpicsSignal, "stoppable", kind="config")
    status = Cpt(EpicsSignal, "status", kind="config")
    console = Cpt(EpicsSignal, "console", kind="config")

    start_ioc = Cpt(EpicsSignal, "start", kind="omitted")
    stop_ioc = Cpt(EpicsSignal, "stop", kind="omitted")
    restart_ioc = Cpt(EpicsSignal, "restart", kind="omitted")


class BeamlineManager(Device):
    """A beamline manager IOC.

    IOCs
    ----

    The beamline manager can control other IOCs. To properly set this
    up, pass a dictionary called *iocs* when initializing this object,
    which should have the IOC names as keys and the corresponding PV
    suffixes as values. This will create a dynamic device component
    *iocs* matching the IOC managers.

    """

    bss = Cpt(BSS, "bss:", name="bss")
    local_storage = Cpt(LocalStorage, "local_storage:", name="local_storage")

    def __new__(
        cls: type,
        prefix: str,
        name: str,
        iocs: Mapping = {},
        **kwargs,
    ):
        defn = {key: (IOCManager, val, {}) for key, val in iocs.items()}
        comps = {
            "iocs": DCpt(defn),
        }
        new_cls = type("BeamlineManager", (cls,), comps)
        return object.__new__(new_cls)

    def __init__(
        self,
        prefix: str,
        name: str,
        iocs: Mapping = {},
        labels: set = {"beamline_manager"},
        **kwargs,
    ):
        super().__init__(prefix=prefix, name=name, labels=labels, **kwargs)



"""
Provide beamline configuration from the iconfig.toml file.

Example TOML configuration file: iconfig_default.toml

"""

__all__ = [
    "load_config",
]

import argparse
import logging
import os
from contextlib import contextmanager
from copy import deepcopy
from pathlib import Path
from pprint import pprint
from typing import Optional, Sequence

import tomli
from mergedeep import merge

log = logging.getLogger(__name__)


_local_overrides = {}


def load_files(file_paths: Sequence[Path]):
    """Generate the configs for files as dictionaries."""
    for fp in file_paths:
        fp = Path(fp)
        if fp.exists():
            with open(fp, mode="rb") as fp:
                log.debug(f"Loading config file: {fp}")
                config = tomli.load(fp)
                yield config
        else:
            log.debug(f"Could not find config file, skipping: {fp}")


def lookup_file_paths():
    if os.environ.get("HAVEN_CONFIG_FILES", "") != "":
        return [Path(fp) for fp in os.environ["HAVEN_CONFIG_FILES"].split(",")]
    else:
        return [Path(__file__).parent / "iconfig_default.toml"]


def load_config(file_paths: Optional[Sequence[Path]] = None):
    """Load TOML config files.

    Will load files specified in the following locations:

    1. *file_paths* argument
    2. The $HAVEN_CONFIG_FILES environmental variable.
    3. iconfig_default.toml file included with Haven.

    """
    if file_paths is None:
        file_paths = lookup_file_paths()
    else:
        file_paths = list(file_paths).copy()
    # Add config file from environmental variable
    # Load configuration from TOML files
    config = {}
    merge(config, *load_files(file_paths), _local_overrides)
    return config


def print_config_value(args: Sequence[str] = None):
    """Print a config value from TOML files.

    Parameters
    ----------
    key
      The path to the value to retrieve from config files. Sections
      should be separated by dots, e.g. "shutter.A.open_pv"

    """
    # Set up command line arguments
    parser = argparse.ArgumentParser(
        prog="haven_config",
        description="Retrieve a value from Haven's config files.",
    )
    parser.add_argument("key", help="The dot-separated key to look up.")
    args = parser.parse_args(args=args)
    # Get the keys from the config file
    value = load_config()
    for part in args.key.split("."):
        value = value[part]
    try:
        value = value.strip()
    except AttributeError:
        # It's not a simple string, so pretty print it
        pprint(value)
    else:
        # Simple string, so just print it
        print(value)


@contextmanager
def beamline_connected(is_connected=True):
    global _local_overrides
    # Save old value
    old_dict = deepcopy(_local_overrides)
    # Set temporary value
    if "beamline" not in _local_overrides.keys():
        _local_overrides["beamline"] = {}
    _local_overrides["beamline"]["hardware_is_present"] = is_connected
    # Return to enclosing code
    yield
    # Restore old value
    _local_overrides = old_dict


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import time
from collections import OrderedDict

import pytest
from ophyd import ADBase
from ophyd import ADComponent as ADCpt
from ophyd.areadetector.cam import AreaDetectorCam
from ophyd.sim import instantiate_fake_device

from haven.devices.area_detector import (
    DetectorBase,
    DetectorState,
    HDF5FilePlugin,
    make_area_detector,
)


class Detector(DetectorBase):
    cam = ADCpt(AreaDetectorCam, "cam1:")
    hdf = ADCpt(HDF5FilePlugin, "HDF1:", write_path_template="/tmp/")


@pytest.fixture()
def detector(sim_registry):
    det = instantiate_fake_device(Detector)
    return det


def test_flyscan_kickoff(detector):
    detector.flyer_num_points.set(10)
    status = detector.kickoff()
    detector.cam.detector_state.sim_put(DetectorState.ACQUIRE)
    status.wait(timeout=3)
    assert status.success
    assert status.done
    # Check that the device was properly configured for fly-scanning
    assert detector.cam.acquire.get() == 1
    assert detector._fly_data == {}
    # Check that timestamps get recorded when new data are available
    detector.cam.array_counter.sim_put(1)
    event = detector._fly_data[detector.cam.array_counter]
    assert event[0].timestamp == pytest.approx(time.time())


def test_make_area_detector(sim_registry):
    ad = make_area_detector(
        name="ad_sim", prefix="255idADSimDet:", device_class="SimDetector", mock=True
    )
    # Check that some area detectors were loaded
    assert isinstance(ad, ADBase)
    # dets = sim_registry.findall(label="area_detectors")


def test_hdf_dtype(detector):
    """Check that the right ``dtype_str`` is added to the image data to
    make tiled happy.
    """
    # Set up fake image metadata
    detector.hdf.data_type.sim_put("UInt8")
    original_desc = OrderedDict(
        [
            (
                "FakeDetector_image",
                {
                    "shape": (1, 1024, 1280),
                    "source": "PV:25idcARV4:",
                    "dtype": "array",
                    "external": "FILESTORE:",
                },
            )
        ]
    )
    # Update and check the description
    new_desc = detector.hdf._add_dtype_str(original_desc)
    assert new_desc["FakeDetector_image"]["dtype_str"] == "|u1"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import os
import subprocess
from pathlib import Path

from haven import load_config


def launch_queueserver():
    # Derive environmental variables
    this_dir = Path(__file__).parent
    bluesky_dir = Path(os.environ["BLUESKY_DIR"])
    default_config_file = str(bluesky_dir / "iconfig.toml")
    config_files = os.environ.setdefault(
        "HAVEN_CONFIG_FILES", default_config_file
    ).split(",")
    # Derive internal haven variables

    config = load_config(file_paths=config_files)
    control_port = config["queueserver"]["control_port"]
    info_port = config["queueserver"]["info_port"]
    kafka_topic = config["queueserver"]["kafka_topic"]
    redis_addr = config["queueserver"]["redis_addr"]
    # Launch the queueserver
    args = [
        "start-re-manager",
        "--startup-script",
        str(this_dir / "queueserver_startup.py"),
        "--existing-plans-devices",
        str(bluesky_dir / "queueserver_existing_plans_and_devices.yaml"),
        "--user-group-permissions",
        str(this_dir / "queueserver_user_group_permissions.yaml"),
        "--zmq-control-addr",
        f"tcp://*:{control_port}",
        "--zmq-info-addr",
        f"tcp://*:{info_port}",
        "--redis-addr",
        redis_addr,
        "--keep-re",
        "--kafka-topic",
        kafka_topic,
        "--update-existing-plans-devices",
        "ENVIRONMENT_OPEN",
        "--use-ipython-kernel=ON",
    ]
    print("Starting queueserver with command:")
    print("  ", " ".join(args))
    subprocess.run(args)



import logging
import sys

import numpy as np
import pydm
import pyqtgraph

from firefly import display
from haven import beamline

np.set_printoptions(threshold=sys.maxsize)


log = logging.getLogger(__name__)


pyqtgraph.setConfigOption("imageAxisOrder", "row-major")


class AreaDetectorViewerDisplay(display.FireflyDisplay):
    caqtdm_ui_file: str = (
        "/APSshare/epics/synApps_6_2_1/support/areaDetector-R3-12-1/ADAravis/aravisApp/op/ui/autoconvert/ADAravis.ui"
    )
    image_is_new: bool = True

    def customize_device(self):
        device_name = name = self.macros()["AD"]
        device = beamline.registry.find(device_name)
        self.device = device
        img_pv = device.pva.pv_name.get(as_string=True)
        addr = f"pva://{img_pv}"
        self.image_channel = pydm.PyDMChannel(
            address=addr, value_slot=self.update_image
        )
        self.image_channel.connect()

    def customize_ui(self):
        # Create the pyqtgraph image viewer
        self.image_view = self.ui.image_view
        # Connect signals for showing/hiding controls
        self.ui.settings_button.clicked.connect(self.toggle_controls)
        # Set some text about the camera
        use_name = getattr(self.device, "description", None) in [self.device.name, None]
        if use_name:
            lbl_text = self.device.cam.name
        else:
            lbl_text = f"{self.device.description} ({self.device.cam.prefix})"
        self.ui.camera_description_label.setText(lbl_text)
        self.setWindowTitle(lbl_text)

    def toggle_controls(self):
        # Show or hide the controls frame
        controls_frame = self.ui.controls_frame
        controls_frame.setVisible(not controls_frame.isVisible())

    def update_image(self, img):
        # For some reason the image comes out mishapen
        new_shape = img.shape[::-1]
        img = np.reshape(img, new_shape)
        # Show the image data
        self.image_view.setImage(
            img, autoRange=self.image_is_new, autoLevels=self.image_is_new
        )
        # Update the display to indicate that we don't need to update levels/scale in the future
        self.image_is_new = False

    def ui_filename(self):
        return "area_detector_viewer.ui"

    def launch_caqtdm(self):
        # Determine for which IOC to launch caQtDM panels
        cmd = f"start_{self.device.prefix.strip(':')}_caqtdm"
        # Launch caQtDM for the given IOC
        log.info(f"Launching caQtDM: {cmd}")
        self._open_caqtdm_subprocess(cmd)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



"""Tests to check that the simulated tiled client works properly."""

import numpy
import pytest
from tiled.adapters.array import ArrayAdapter
from tiled.adapters.mapping import MapAdapter
from tiled.client import Context, from_context
from tiled.server.app import build_app

simple_tree = MapAdapter(
    {
        "a": ArrayAdapter.from_array(
            numpy.arange(10), metadata={"apple": "red", "animal": "dog"}
        ),
        "b": ArrayAdapter.from_array(
            numpy.arange(10), metadata={"banana": "yellow", "animal": "dog"}
        ),
        "c": ArrayAdapter.from_array(
            numpy.arange(10), metadata={"cantalope": "orange", "animal": "cat"}
        ),
    }
)


@pytest.fixture(scope="module")
def client():
    app = build_app(simple_tree)
    with Context.from_app(app) as context:
        client = from_context(context)
        yield client


def test_client_fixture(client):
    """Does the client fixture load without stalling the test runner?"""


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import subprocess

import psutil
import pydm
import pytest
from ophyd.sim import make_fake_device
from ophyd_async.core import wait_for_connection

from haven.devices.motor import HavenMotor, Motor


def tiled_is_running(port, match_command=True):
    lsof = subprocess.run(["lsof", "-i", f":{port}", "-F"], capture_output=True)
    assert lsof.stderr.decode() == ""
    stdout = lsof.stdout.decode().split("\n")
    is_running = len(stdout) >= 3
    if match_command:
        is_running = is_running and stdout[3] == "ctiled"
    return is_running


def kill_process(process_name):
    processes = []
    for proc in psutil.process_iter():
        # check whether the process name matches
        if proc.name() == process_name:
            proc.kill()
            processes.append(proc)
    # Wait for them all the terminate
    [proc.wait(timeout=5) for proc in processes]


@pytest.fixture(scope="session")
def pydm_ophyd_plugin():
    return pydm.data_plugins.plugin_for_address("haven://")


qs_status = {
    "msg": "RE Manager v0.0.18",
    "items_in_queue": 0,
    "items_in_history": 0,
    "running_item_uid": None,
    "manager_state": "idle",
    "queue_stop_pending": False,
    "worker_environment_exists": False,
    "worker_environment_state": "closed",
    "worker_background_tasks": 0,
    "re_state": None,
    "pause_pending": False,
    "run_list_uid": "4f2d48cc-980d-4472-b62b-6686caeb3833",
    "plan_queue_uid": "2b99ccd8-f69b-4a44-82d0-947d32c5d0a2",
    "plan_history_uid": "9af8e898-0f00-4e7a-8d97-0964c8d43f47",
    "devices_existing_uid": "51d8b88d-7457-42c4-b67f-097b168be96d",
    "plans_existing_uid": "65f11f60-0049-46f5-9eb3-9f1589c4a6dd",
    "devices_allowed_uid": "a5ddff29-917c-462e-ba66-399777d2442a",
    "plans_allowed_uid": "d1e907cd-cb92-4d68-baab-fe195754827e",
    "plan_queue_mode": {"loop": False},
    "task_results_uid": "159e1820-32be-4e01-ab03-e3478d12d288",
    "lock_info_uid": "c7fe6f73-91fc-457d-8db0-dfcecb2f2aba",
    "lock": {"environment": False, "queue": False},
}


@pytest.fixture
def sync_motors(sim_registry):
    motor_names = ["sync_motor_1", "sync_motor_2"]
    motors = []
    for name in motor_names:
        this_motor = make_fake_device(HavenMotor)(name=name, labels={"motors"})
        motors.append(this_motor)
        sim_registry.register(this_motor)
    return motors


@pytest.fixture
async def async_motors(sim_registry):
    motors = [
        Motor(prefix="255idc:m1", name="async_motor_1"),
        Motor(prefix="255idc:m2", name="async_motor_2"),
    ]
    await wait_for_connection(**{m.name: m.connect(mock=True) for m in motors})
    for motor in motors:
        sim_registry.register(motor)
    return motors
